#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import resource
import pytest
import tempfile
import typing as tp
import uuid as sys_uuid
from types import SimpleNamespace
from unittest import mock

from gcl_sdk.agents.universal.dm import models as ua_models
from gcl_sdk.agents.universal.services import builder as builder_svc
from gcl_sdk.agents.universal import constants as ua_c
from gcl_sdk.tests.functional import conftest
from gcl_sdk.tests.functional import utils as test_utils


class DummyBuilder(builder_svc.UniversalBuilderService):
    def __init__(self, instance_model):
        super().__init__(instance_model=instance_model)
        self.pre_create_called = False
        self.post_create_called = False
        self.pre_update_called = False
        self.post_update_called = False
        self.create_instance_derivatives_called = False
        self.deleted_pre_called = []

    def pre_create_instance_resource(self, instance):
        self.pre_create_called = True

    def post_create_instance_resource(
        self, instance, resource, derivatives=tuple()
    ):
        self.post_create_called = True
        instance.status = ua_c.InstanceStatus.IN_PROGRESS.value

    def pre_update_instance_resource(self, instance):
        self.pre_update_called = True
        instance.status = ua_c.InstanceStatus.IN_PROGRESS.value

    def post_update_instance_resource(self, instance, resource):
        self.post_update_called = True

    def pre_delete_instance_resource(self, resource):
        self.deleted_pre_called.append(resource)

    def create_instance_derivatives(
        self, instance: ua_models.InstanceMixin
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        self.create_instance_derivatives_called = True
        return tuple()


class DummyBuilderWithDerivatives(DummyBuilder):
    def create_instance_derivatives(
        self, instance: conftest.DummyInstanceWithDerivatives
    ) -> tp.Collection[conftest.DummyDerivative]:
        self.create_instance_derivatives_called = True
        return [instance.to_derivative()]


class DummyBuilderWithMasterTracking(DummyBuilderWithDerivatives):
    def track_outdated_master_hash_instances(self) -> bool:
        """Track outdated master hash instances."""
        return True

    def track_outdated_master_full_hash_instances(self) -> bool:
        """Track outdated master full hash instances."""
        return True

    def actualize_outdated_master_hash_instance(
        self,
        instance,
        master_instance,
        derivatives,
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        return tuple(p[0] for p in derivatives)

    def actualize_outdated_master_full_hash_instance(
        self,
        instance: ua_models.InstanceMixin,
        master_instance: ua_models.InstanceMixin,
        derivatives: tp.Collection[
            tuple[
                ua_models.TargetResourceKindAwareMixin,  # The target resource
                ua_models.TargetResourceKindAwareMixin
                | None,  # The actual resource
            ]
        ],
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        return tuple(p[0] for p in derivatives)


class TestUniversalBuilderService:

    # Need only to apply DB migrations
    @pytest.fixture(scope="class", autouse=True)
    def api_service(self, orch_api_wsgi_app):
        class ApiRestService(test_utils.RestServiceTestCase):
            __FIRST_MIGRATION__ = conftest.FIRST_MIGRATION
            __APP__ = orch_api_wsgi_app

        rest_service = ApiRestService()
        rest_service.setup_class()

        yield rest_service

        rest_service.teardown_class()

    @pytest.fixture(autouse=True)
    def db_migrations(self, api_service: test_utils.RestServiceTestCase):
        api_service.setup_method()
        yield api_service
        api_service.teardown_method()

    def test_no_actions(self):
        svc = DummyBuilder(instance_model=conftest.DummyInstance)
        svc._iteration()

    # No derivatives

    def test_create_new_instance(self, dummy_instance_factory: tp.Callable):
        instance = dummy_instance_factory({"new": 1})[0]

        svc = DummyBuilder(instance_model=conftest.DummyInstance)
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 1

        resource = resources[0]
        assert resource.uuid == instance.uuid
        assert resource.kind == conftest.DummyInstance.get_resource_kind()

        resources = ua_models.Resource.objects.get_all()
        assert len(resources) == 0

        assert svc.pre_create_called
        assert svc.post_create_called
        assert not svc.create_instance_derivatives_called

    def test_create_multiple_instances(
        self, dummy_instance_factory: tp.Callable
    ):
        instances = dummy_instance_factory({"new": 2})

        svc = DummyBuilder(instance_model=conftest.DummyInstance)
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == len(instances)

        resource = resources[0]
        assert {r.uuid for r in resources} == {i.uuid for i in instances}
        assert resource.kind == conftest.DummyInstance.get_resource_kind()

        resources = ua_models.Resource.objects.get_all()
        assert len(resources) == 0

        assert svc.pre_create_called
        assert svc.post_create_called
        assert not svc.create_instance_derivatives_called

    def test_delete_instance(self, dummy_instance_factory: tp.Callable):
        resource = dummy_instance_factory({"deleted": 1})[0]

        svc = DummyBuilder(instance_model=conftest.DummyInstance)
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 0

        resources = ua_models.Resource.objects.get_all()
        assert len(resources) == 0

        assert svc.deleted_pre_called[0] == resource
        assert not svc.pre_create_called
        assert not svc.post_create_called
        assert not svc.create_instance_derivatives_called

    def test_delete_multiple_instances(
        self, dummy_instance_factory: tp.Callable
    ):
        instance_resources = dummy_instance_factory({"deleted": 2})

        svc = DummyBuilder(instance_model=conftest.DummyInstance)
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 0

        resources = ua_models.Resource.objects.get_all()
        assert len(resources) == 0

        assert svc.deleted_pre_called == instance_resources
        assert not svc.pre_create_called
        assert not svc.post_create_called
        assert not svc.create_instance_derivatives_called

    def test_actualize_updated_instances(
        self, dummy_instance_factory: tp.Callable
    ):
        instance = dummy_instance_factory({"updated": 1})[0]

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 1
        resource = resources[0]
        assert resource.uuid == instance.uuid
        assert resource.value["name"] != instance.name

        svc = DummyBuilder(instance_model=conftest.DummyInstance)
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 1

        resource = resources[0]
        assert resource.uuid == instance.uuid
        assert resource.value["name"] == instance.name
        assert resource.kind == conftest.DummyInstance.get_resource_kind()

        resources = ua_models.Resource.objects.get_all()
        assert len(resources) == 0

        assert not svc.pre_create_called
        assert not svc.post_create_called
        assert not svc.create_instance_derivatives_called
        assert svc.pre_update_called
        assert svc.post_update_called

    def test_actualize_updated_multiple_instances(
        self, dummy_instance_factory: tp.Callable
    ):
        instances = dummy_instance_factory({"updated": 2})

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 2
        assert {r.uuid for r in resources} == {i.uuid for i in instances}
        assert not (
            {r.value["name"] for r in resources} & {i.name for i in instances}
        )

        svc = DummyBuilder(instance_model=conftest.DummyInstance)
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 2
        assert {r.uuid for r in resources} == {i.uuid for i in instances}
        assert {r.value["name"] for r in resources} == {
            i.name for i in instances
        }

        resources = ua_models.Resource.objects.get_all()
        assert len(resources) == 0

        assert not svc.pre_create_called
        assert not svc.post_create_called
        assert not svc.create_instance_derivatives_called
        assert svc.pre_update_called
        assert svc.post_update_called

    def test_actualize_outdated_instances(
        self, dummy_instance_factory: tp.Callable
    ):
        class Builder(DummyBuilder):
            actualized = False

            def actualize_outdated_instance(
                self,
                current_instance,
                actual_instance,
            ) -> None:
                self.__class__.actualized = True

        # Prepare: create an existing target resource (no derivatives)
        instance = dummy_instance_factory({"existing": 1})[0]

        conftest.DummyInstance.objects = mock.MagicMock()
        conftest.DummyInstance.objects.get_all = mock.MagicMock(
            return_value=[instance]
        )

        # Sanity: only target resource exists, no actual resources yet
        t_resources = ua_models.TargetResource.objects.get_all()
        a_resources = ua_models.Resource.objects.get_all()
        assert len(t_resources) == 1
        assert len(a_resources) == 0

        # Create an actual resource with a different value/status to mark it as outdated
        target = t_resources[0]
        assert target.kind == conftest.DummyInstance.get_resource_kind()

        actual_value = instance.dump_to_simple_view()
        actual = ua_models.Resource.from_value(
            value=actual_value,
            kind=target.kind,
        )
        actual.save()

        # Run actualization
        svc = Builder(instance_model=conftest.DummyInstance)
        svc._iteration()

        # Validate target resource is actualized from actual (status/full_hash)
        t_resources = ua_models.TargetResource.objects.get_all()
        a_resources = ua_models.Resource.objects.get_all()
        assert len(t_resources) == 1
        assert len(a_resources) == 1

        target = t_resources[0]
        actual = a_resources[0]

        assert target.uuid == actual.uuid
        assert target.kind == actual.kind
        # full_hash is synced with actual; status follows actual
        assert target.full_hash == actual.full_hash
        assert target.status == actual.status

        assert Builder.actualized

    def test_actualize_outdated_instances_failed(
        self, dummy_instance_factory: tp.Callable
    ):
        class Builder(DummyBuilder):
            called = 0

            def actualize_outdated_instance(
                self,
                current_instance,
                actual_instance,
            ) -> None:
                self.__class__.called += 1
                raise RuntimeError("Failed to actualize instance")

        # Prepare: create an existing target resource (no derivatives)
        instance = dummy_instance_factory({"existing": 1})[0]

        conftest.DummyInstance.objects = mock.MagicMock()
        conftest.DummyInstance.objects.get_all = mock.MagicMock(
            return_value=[instance]
        )

        # Sanity: only target resource exists, no actual resources yet
        t_resources = ua_models.TargetResource.objects.get_all()
        a_resources = ua_models.Resource.objects.get_all()
        assert len(t_resources) == 1
        assert len(a_resources) == 0

        # Create an actual resource with a different value/status to mark it as outdated
        target = t_resources[0]
        assert target.kind == conftest.DummyInstance.get_resource_kind()

        actual_value = instance.dump_to_simple_view()
        actual = ua_models.Resource.from_value(
            value=actual_value,
            kind=target.kind,
        )
        actual.save()

        # Run actualization
        svc = Builder(instance_model=conftest.DummyInstance)
        svc._iteration()

        # Validate target resource is actualized from actual (status/full_hash)
        t_resources = ua_models.TargetResource.objects.get_all()
        a_resources = ua_models.Resource.objects.get_all()
        assert len(t_resources) == 1
        assert len(a_resources) == 1

        target = t_resources[0]
        actual = a_resources[0]

        assert target.uuid == actual.uuid
        assert target.kind == actual.kind
        # full_hash is synced with actual; status follows actual
        assert target.full_hash != actual.full_hash

        assert Builder.called == 1

        svc._iteration()

        t_resources = ua_models.TargetResource.objects.get_all()
        a_resources = ua_models.Resource.objects.get_all()
        assert len(t_resources) == 1
        assert len(a_resources) == 1

        target = t_resources[0]
        actual = a_resources[0]

        assert target.uuid == actual.uuid
        assert target.kind == actual.kind
        # full_hash is synced with actual; status follows actual
        assert target.full_hash != actual.full_hash
        assert Builder.called == 2

    # Derivatives

    def test_create_new_instance_derivatives(
        self, dummy_instance_with_derivatives_factory: tp.Callable
    ):
        instance = dummy_instance_with_derivatives_factory({"new": 1})[0]

        svc = DummyBuilderWithDerivatives(
            instance_model=conftest.DummyInstanceWithDerivatives
        )
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 2
        assert {r.kind for r in resources} == {"foo", "foo-derivative"}
        assert {r.uuid for r in resources} == {
            instance.uuid,
            sys_uuid.uuid5(instance.uuid, instance.name),
        }

        resources = ua_models.Resource.objects.get_all()
        assert len(resources) == 0

        assert svc.pre_create_called
        assert svc.post_create_called
        assert svc.create_instance_derivatives_called

    def test_create_multiple_instance_derivatives(
        self, dummy_instance_with_derivatives_factory: tp.Callable
    ):
        instances = dummy_instance_with_derivatives_factory({"new": 2})

        svc = DummyBuilderWithDerivatives(
            instance_model=conftest.DummyInstanceWithDerivatives
        )
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 4

        assert {r.kind for r in resources} == {"foo", "foo-derivative"}

        resources = ua_models.Resource.objects.get_all()
        assert len(resources) == 0

        assert svc.pre_create_called
        assert svc.post_create_called
        assert svc.create_instance_derivatives_called

    def test_delete_instance_derivatives(
        self, dummy_instance_with_derivatives_factory: tp.Callable
    ):
        resource = dummy_instance_with_derivatives_factory({"deleted": 1})[0]

        svc = DummyBuilderWithDerivatives(
            instance_model=conftest.DummyInstanceWithDerivatives
        )
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 0

        resources = ua_models.Resource.objects.get_all()
        assert len(resources) == 0

        assert svc.deleted_pre_called[0] == resource
        assert not svc.pre_create_called
        assert not svc.post_create_called
        assert not svc.create_instance_derivatives_called

    def test_actualize_updated_instance_derivatives(
        self, dummy_instance_with_derivatives_factory: tp.Callable
    ):
        instance = dummy_instance_with_derivatives_factory({"updated": 1})[0]

        resources = ua_models.TargetResource.objects.get_all()
        derivative = [r for r in resources if r.master == instance.uuid][0]

        assert len(resources) == 2
        assert derivative.master == instance.uuid
        assert {r.uuid for r in resources} == {
            instance.uuid,
            sys_uuid.uuid5(instance.uuid, "inst-updated-0"),
        }
        assert not ({r.value["name"] for r in resources} & {instance.name})

        svc = DummyBuilderWithDerivatives(
            instance_model=conftest.DummyInstanceWithDerivatives
        )
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 2

        assert {r.uuid for r in resources} == {
            instance.uuid,
            sys_uuid.uuid5(instance.uuid, instance.name),
        }
        assert {r.value["name"] for r in resources} == {instance.name}

        resources = ua_models.Resource.objects.get_all()
        assert len(resources) == 0

        assert not svc.pre_create_called
        assert not svc.post_create_called
        assert svc.create_instance_derivatives_called
        assert svc.pre_update_called
        assert svc.post_update_called

    def test_actualize_outdated_instances_derivatives(
        self, dummy_instance_with_derivatives_factory: tp.Callable
    ):
        class Builder(DummyBuilderWithDerivatives):
            actualized = False

            def actualize_outdated_instance_derivatives(
                self,
                instance,
                derivative_pairs,
            ):
                self.__class__.actualized = True
                assert len(derivative_pairs) == 1
                assert instance.get_resource_kind() == "foo"
                assert (
                    derivative_pairs[0][0].get_resource_kind()
                    == "foo-derivative"
                )

                return tuple(p[0] for p in derivative_pairs)

        # Prepare: create an existing target resource (no derivatives)
        instance = dummy_instance_with_derivatives_factory({"existing": 1})[0]

        conftest.DummyInstanceWithDerivatives.objects = mock.MagicMock()
        conftest.DummyInstanceWithDerivatives.objects.get_all = mock.MagicMock(
            return_value=[instance]
        )

        # Sanity: only target resource exists, no actual resources yet
        t_resources = ua_models.TargetResource.objects.get_all()
        a_resources = ua_models.Resource.objects.get_all()
        assert len(t_resources) == 2
        assert len(a_resources) == 0

        derivative = instance.to_derivative()
        actual_value = derivative.dump_to_simple_view()
        actual = ua_models.Resource.from_value(
            value=actual_value,
            kind=derivative.get_resource_kind(),
        )
        actual.save()

        # Run actualization
        svc = Builder(instance_model=conftest.DummyInstanceWithDerivatives)
        svc._iteration()

        # Validate target resource is actualized from actual (status/full_hash)
        t_resources = ua_models.TargetResource.objects.get_all()
        a_resources = ua_models.Resource.objects.get_all()
        assert len(t_resources) == 2
        assert len(a_resources) == 1

        target = [r for r in t_resources if r.kind == "foo-derivative"][0]
        actual = a_resources[0]

        assert target.uuid == actual.uuid
        assert target.kind == actual.kind
        # full_hash is synced with actual; status follows actual
        assert target.full_hash == actual.full_hash

        assert Builder.actualized

    def test_actualize_outdated_instances_derivatives_failed(
        self, dummy_instance_with_derivatives_factory: tp.Callable
    ):
        class Builder(DummyBuilderWithDerivatives):
            called = 0

            def actualize_outdated_instance_derivatives(
                self,
                instance,
                derivative_pairs,
            ):
                self.__class__.called += 1
                assert len(derivative_pairs) == 1
                assert instance.get_resource_kind() == "foo"
                assert (
                    derivative_pairs[0][0].get_resource_kind()
                    == "foo-derivative"
                )

                raise RuntimeError("Failed to actualize instance derivatives")

        # Prepare: create an existing target resource (no derivatives)
        instance = dummy_instance_with_derivatives_factory({"existing": 1})[0]

        conftest.DummyInstanceWithDerivatives.objects = mock.MagicMock()
        conftest.DummyInstanceWithDerivatives.objects.get_all = mock.MagicMock(
            return_value=[instance]
        )

        # Sanity: only target resource exists, no actual resources yet
        t_resources = ua_models.TargetResource.objects.get_all()
        a_resources = ua_models.Resource.objects.get_all()
        assert len(t_resources) == 2
        assert len(a_resources) == 0

        derivative = instance.to_derivative()
        actual_value = derivative.dump_to_simple_view()
        actual = ua_models.Resource.from_value(
            value=actual_value,
            kind=derivative.get_resource_kind(),
        )
        actual.save()

        # Run actualization
        svc = Builder(instance_model=conftest.DummyInstanceWithDerivatives)
        svc._iteration()

        # Validate target resource is actualized from actual (status/full_hash)
        t_resources = ua_models.TargetResource.objects.get_all()
        a_resources = ua_models.Resource.objects.get_all()
        assert len(t_resources) == 2
        assert len(a_resources) == 1

        target = [r for r in t_resources if r.kind == "foo-derivative"][0]
        actual = a_resources[0]

        assert target.uuid == actual.uuid
        assert target.kind == actual.kind
        # full_hash is synced with actual; status follows actual
        assert target.full_hash != actual.full_hash

        assert Builder.called == 1

        svc._iteration()

        t_resources = ua_models.TargetResource.objects.get_all()
        a_resources = ua_models.Resource.objects.get_all()
        target = [r for r in t_resources if r.kind == "foo-derivative"][0]
        actual = a_resources[0]

        assert target.full_hash != actual.full_hash
        assert Builder.called == 2

    # Master tracking

    def test_track_master_hash_instance(
        self, dummy_instance_with_derivatives_factory: tp.Callable
    ):
        class DummyBuilderWithMasterTracking(DummyBuilderWithDerivatives):
            _actualized = False

            def track_outdated_master_hash_instances(self) -> bool:
                """Track outdated master hash instances."""
                return True

            def actualize_outdated_master_hash_instance(
                self,
                instance,
                master_instance,
                derivatives,
            ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
                self.__class__._actualized = True
                master_kind = master_instance.get_resource_kind()
                instance_kind = instance.get_resource_kind()
                assert master_kind == "foo"
                assert instance_kind == "sub-foo"

                return tuple(p[0] for p in derivatives)

        instance = dummy_instance_with_derivatives_factory(
            {"existing": 1}, model=conftest.DummySubInstanceWithDerivatives
        )[0]

        svc = DummyBuilderWithMasterTracking(
            instance_model=conftest.DummySubInstanceWithDerivatives
        )
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 3

        instance_resource = [r for r in resources if r.kind == "sub-foo"][0]
        instance_resource.master_hash = "1111"
        instance_resource.save()

        svc._iteration()

        assert DummyBuilderWithMasterTracking._actualized

        resources = ua_models.TargetResource.objects.get_all()
        master = [r for r in resources if r.kind == "foo"][0]
        instance = [r for r in resources if r.kind == "sub-foo"][0]

        assert master.hash == instance.master_hash
        assert instance.full_hash != "1111"

    def test_track_master_full_hash_instance(
        self, dummy_instance_with_derivatives_factory: tp.Callable
    ):
        class DummyBuilderWithMasterTracking(DummyBuilderWithDerivatives):
            _actualized = False

            def track_outdated_master_full_hash_instances(self) -> bool:
                """Track outdated master hash instances."""
                return True

            def actualize_outdated_master_full_hash_instance(
                self,
                instance,
                master_instance,
                derivatives,
            ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
                self.__class__._actualized = True
                master_kind = master_instance.get_resource_kind()
                instance_kind = instance.get_resource_kind()
                assert master_kind == "foo"
                assert instance_kind == "sub-foo"

                return tuple(p[0] for p in derivatives)

        instance = dummy_instance_with_derivatives_factory(
            {"existing": 1}, model=conftest.DummySubInstanceWithDerivatives
        )[0]

        svc = DummyBuilderWithMasterTracking(
            instance_model=conftest.DummySubInstanceWithDerivatives
        )
        svc._iteration()

        resources = ua_models.TargetResource.objects.get_all()
        assert len(resources) == 3

        instance_resource = [r for r in resources if r.kind == "sub-foo"][0]
        instance_resource.master_full_hash = "1111"
        instance_resource.save()

        svc._iteration()

        assert DummyBuilderWithMasterTracking._actualized

        resources = ua_models.TargetResource.objects.get_all()
        master = [r for r in resources if r.kind == "foo"][0]
        instance = [r for r in resources if r.kind == "sub-foo"][0]

        assert master.full_hash == instance.master_full_hash
        assert instance.full_hash != "1111"



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

from urllib.parse import urljoin
import uuid as sys_uuid

import pytest
import requests
from oslo_config import cfg

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.tests.functional import utils as test_utils
from gcl_sdk.tests.functional import conftest


CONF = cfg.CONF


class TestUAOrchApi:
    @pytest.fixture(scope="class")
    def orch_api_service(self, orch_api_wsgi_app):
        class ApiRestService(test_utils.RestServiceTestCase):
            __FIRST_MIGRATION__ = conftest.FIRST_MIGRATION
            __APP__ = orch_api_wsgi_app

        rest_service = ApiRestService()
        rest_service.setup_class()

        yield rest_service

        rest_service.teardown_class()

    @pytest.fixture()
    def orch_api(self, orch_api_service: test_utils.RestServiceTestCase):
        orch_api_service.setup_method()

        yield orch_api_service

        orch_api_service.teardown_method()

    def test_agent_no_agents(
        self,
        orch_api: test_utils.RestServiceTestCase,
    ):
        url = urljoin(orch_api.base_url, "agents/")

        response = requests.get(url)

        assert response.status_code == 200

    def test_agent_list(
        self,
        orch_api: test_utils.RestServiceTestCase,
    ):
        uuid_a = sys_uuid.uuid4()
        agent_a = models.UniversalAgent(
            name="Agent A", uuid=uuid_a, node=uuid_a
        )
        uuid_b = sys_uuid.uuid4()
        agent_b = models.UniversalAgent(
            name="Agent B", uuid=uuid_b, node=uuid_b
        )
        agent_a.insert()
        agent_b.insert()

        url = urljoin(orch_api.base_url, "agents/")

        response = requests.get(url)
        output = response.json()

        assert response.status_code == 200
        assert len(output) == 2
        assert {a["uuid"] for a in output} == {str(uuid_a), str(uuid_b)}

    def test_agent_register(
        self,
        orch_api: test_utils.RestServiceTestCase,
    ):
        uuid = sys_uuid.uuid4()
        agent = models.UniversalAgent(
            name="Agent A",
            uuid=uuid,
            node=uuid,
            capabilities={"capabilities": ["foo"]},
            facts={"facts": ["bar"]},
        )

        view = agent.dump_to_simple_view()

        url = urljoin(orch_api.base_url, "agents/")

        response = requests.post(url, json=view)
        output = response.json()

        assert response.status_code == 201
        assert output["uuid"] == str(uuid)

    def test_agent_get_payload_no_payload(
        self,
        orch_api: test_utils.RestServiceTestCase,
    ):
        uuid = sys_uuid.uuid4()
        agent = models.UniversalAgent(
            name="Agent A",
            uuid=uuid,
            node=uuid,
            capabilities={"capabilities": ["foo"]},
            facts={"facts": ["bar"]},
        )

        view = agent.dump_to_simple_view()

        url = urljoin(orch_api.base_url, "agents/")

        response = requests.post(url, json=view)
        assert response.status_code == 201

        url = urljoin(
            orch_api.base_url,
            f"agents/{agent.uuid}/actions/get_payload?version=0&hash=",
        )

        response = requests.get(url)
        output = response.json()

        assert response.status_code == 200
        assert output["capabilities"] == {"foo": {"resources": []}}
        assert output["facts"] == {
            "bar": {"resources": []},
            "foo": {"resources": []},
        }
        assert output["hash"] != ""

    def test_agent_get_payload(
        self,
        orch_api: test_utils.RestServiceTestCase,
    ):
        uuid = sys_uuid.uuid4()
        agent = models.UniversalAgent(
            name="Agent A",
            uuid=uuid,
            node=uuid,
            capabilities={"capabilities": ["foo"]},
            facts={"facts": ["bar"]},
        )

        view = agent.dump_to_simple_view()

        url = urljoin(orch_api.base_url, "agents/")

        response = requests.post(url, json=view)
        assert response.status_code == 201

        uuid_r = sys_uuid.uuid4()
        resource = conftest.FooTargetResource(
            uuid=uuid_r,
            name="foo-name",
            project_id=uuid_r,
        ).to_ua_resource(kind="foo")
        resource.agent = agent.uuid
        resource.insert()

        url = urljoin(
            orch_api.base_url,
            f"agents/{agent.uuid}/actions/get_payload?version=0&hash=",
        )

        response = requests.get(url)
        output = response.json()

        assert response.status_code == 200

        view = resource.dump_to_simple_view()
        view.pop("agent", None)
        view.pop("master", None)
        view.pop("node", None)
        view.pop("tracked_at", None)
        view.pop("master_hash", None)
        view.pop("master_full_hash", None)
        assert output["capabilities"]["foo"]["resources"] == [view]
        assert output["facts"] == {
            "bar": {"resources": []},
            "foo": {"resources": []},
        }
        assert output["hash"] != ""

    def test_agent_get_payload_same_hash(
        self,
        orch_api: test_utils.RestServiceTestCase,
    ):
        uuid = sys_uuid.uuid4()
        agent = models.UniversalAgent(
            name="Agent A",
            uuid=uuid,
            node=uuid,
            capabilities={"capabilities": ["foo"]},
            facts={"facts": ["bar"]},
        )

        view = agent.dump_to_simple_view()

        url = urljoin(orch_api.base_url, "agents/")

        response = requests.post(url, json=view)
        assert response.status_code == 201

        uuid_r = sys_uuid.uuid4()
        resource = conftest.FooTargetResource(
            uuid=uuid_r,
            name="foo-name",
            project_id=uuid_r,
        ).to_ua_resource(kind="foo")
        resource.agent = agent.uuid
        resource.insert()

        url = urljoin(
            orch_api.base_url,
            f"agents/{agent.uuid}/actions/get_payload?version=0&hash=",
        )

        response = requests.get(url)
        output = response.json()

        assert response.status_code == 200

        current_hash = output["hash"]

        url = urljoin(
            orch_api.base_url,
            (
                f"agents/{agent.uuid}/actions/get_payload?"
                f"version=0&hash={current_hash}"
            ),
        )

        response = requests.get(url)
        output = response.json()

        assert response.status_code == 200

        assert output["capabilities"] == {}
        assert output["facts"] == {}
        assert output["hash"] == current_hash



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import logging
import re
import typing as tp
import uuid as sys_uuid

from gcl_looper.services import basic as looper_basic

from gcl_sdk.agents.universal.drivers import base as driver_base
from gcl_sdk.agents.universal.drivers import meta as driver_meta
from gcl_sdk.agents.universal.drivers import exceptions as driver_exc
from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal.clients.orch import base as orch_base
from gcl_sdk.agents.universal.clients.orch import exceptions as orch_exc
from gcl_sdk.agents.universal import constants as c
from gcl_sdk.agents.universal import utils


LOG = logging.getLogger(__name__)


class UniversalAgentService(looper_basic.BasicService):

    def __init__(
        self,
        agent_uuid: sys_uuid.UUID,
        orch_client: orch_base.AbstractOrchClient,
        caps_drivers: list[driver_base.AbstractCapabilityDriver],
        facts_drivers: list[driver_base.AbstractFactDriver],
        payload_path: str | None = c.PAYLOAD_PATH,
        iter_min_period: float = 3,
        iter_pause: float = 0.1,
    ):
        super().__init__(iter_min_period, iter_pause)
        self._orch_client = orch_client
        self._payload_path = payload_path
        self._agent_uuid = agent_uuid

        # Make driver maps by capability and facts:
        # For instance, {"config": ConfigDriver, "service": ServiceDriver}
        self._caps_driver_map = {}
        for d in caps_drivers:
            for capability in d.get_capabilities():
                self._caps_driver_map[capability] = d

        self._facts_driver_map = {}
        for d in facts_drivers:
            for fact in d.get_facts():
                self._facts_driver_map[fact] = d

    def _register_agent(self) -> None:
        capabilities = self._caps_driver_map.keys()
        facts = self._facts_driver_map.keys()
        agent = models.UniversalAgent.from_system_uuid(
            capabilities, facts, self._agent_uuid
        )
        try:
            self._orch_client.agents_create(agent)
            LOG.info("Agent registered: %s", agent.uuid)
        except orch_exc.AgentAlreadyExists:
            LOG.warning("Agent already registered: %s", agent.uuid)

    def _create_resource(
        self,
        driver: driver_base.AbstractCapabilityDriver,
        resource: models.Resource,
    ) -> models.Resource:
        try:
            dp_resource = driver.create(resource)
        except driver_exc.ResourceAlreadyExists:
            LOG.warning("The resource already exists: %s", resource.uuid)
            dp_resource = driver.get(resource)

        return dp_resource

    def _delete_resource(
        self,
        driver: driver_base.AbstractCapabilityDriver,
        resource: models.Resource,
    ) -> None:
        try:
            driver.delete(resource)
        except driver_exc.ResourceNotFound:
            LOG.warning("The resource does not exist: %s", resource.uuid)

    def _update_resource(
        self,
        driver: driver_base.AbstractCapabilityDriver,
        resource: models.Resource,
    ) -> models.Resource:
        dp_resource = driver.update(resource)
        return dp_resource

    def _actualize_resources(
        self,
        driver: driver_base.AbstractCapabilityDriver,
        capability: str,
        resources: list[models.Resource],
    ) -> list[models.Resource]:
        """
        Actualize resources and return a list of resources from data plane.
        """
        try:
            # Perform some preparations for the driver
            driver.start()
            return self._driver_cap_iteration(driver, capability, resources)
        finally:
            # Finalize the driver
            driver.finalize()

    def _driver_cap_iteration(
        self,
        driver: driver_base.AbstractCapabilityDriver,
        capability: str,
        resources: list[models.Resource],
    ) -> list[models.Resource]:
        """
        Actualize resources and return a list of resources from data plane.
        """
        target_resources = {r: r for r in resources}
        actual_resources = {r: r for r in driver.list(capability)}

        # A list of resources to be collected from the data plane
        # for this capability
        collected_resources = []

        # Create all new resources
        for r in target_resources.keys() - actual_resources.keys():
            try:
                resource = self._create_resource(driver, r)
                collected_resources.append(resource)
            except Exception:
                LOG.exception("Error creating resource %s", r.uuid)

        # Delete outdated resources
        for r in actual_resources.keys() - target_resources.keys():
            try:
                self._delete_resource(driver, r)
            except Exception:
                # The resource wasn't deleted so add it back
                collected_resources.append(r)
                LOG.exception("Error deleting resource %s", r.uuid)

        for r in target_resources.keys() & actual_resources.keys():
            # set does not guarantee which instance will be given on
            # intersection therefore get actual and target resources
            # explicitly.
            target_resource = target_resources[r]
            actual_resource = actual_resources[r]

            # Nothing to do if the resources are the same
            if target_resource.hash == actual_resource.hash:
                collected_resources.append(actual_resource)
                continue

            try:
                resource = self._update_resource(driver, target_resource)
                collected_resources.append(resource)
            except Exception:
                LOG.exception("Error updating resource %s", r.uuid)

        return collected_resources

    def _actualize_resource_facts(
        self,
        target_facts: list[dict[str : tp.Any]],
        actual_facts: list[dict[str : tp.Any]],
    ) -> None:
        """Actualize facts in Status API.

        target_facts - The facts collected from the data plane.
        actual_facts - The facts collected from the Status API.
        """
        target_resources = {r["uuid"]: r for r in target_facts}
        actual_resources = {r["uuid"]: r for r in actual_facts}

        for uuid in target_resources.keys() - actual_resources.keys():
            resource = target_resources[uuid]
            resource["node"] = str(utils.node_uuid())
            try:
                resource = models.Resource.restore_from_simple_view(**resource)
                self._orch_client.resources_create(resource)
            except Exception:
                LOG.exception("Error creating resource %s", uuid)

        for uuid in actual_resources.keys() - target_resources.keys():
            resource = actual_resources[uuid]

            try:
                resource = models.Resource.restore_from_simple_view(**resource)
                self._orch_client.resources_delete(resource)
            except Exception:
                LOG.exception("Error deleting resource %s", uuid)

        for uuid in target_resources.keys() & actual_resources.keys():
            target_resource = target_resources[uuid]
            actual_resource = actual_resources[uuid]

            # Nothing to do if the resources are the same
            if target_resource["full_hash"] == actual_resource["full_hash"]:
                continue

            # Remove read-only fields
            resource = target_resource.copy()
            resource.pop("created_at", None)
            resource.pop("updated_at", None)
            resource.pop("res_uuid", None)
            resource.pop("node", None)

            res_uuid = resource.pop("uuid")
            kind = resource.pop("kind")

            try:
                self._orch_client.resources_update(kind, res_uuid, **resource)
            except Exception:
                LOG.exception("Error updating resource %s", uuid)

    def _actualize_facts(self, target_facts: dict, actual_facts: dict) -> None:
        """Actualize facts in Status API.

        target_facts - The facts collected from the data plane.
        actual_facts - The facts collected from the Status API.
        """
        # New fact category
        for fact in target_facts.keys() - actual_facts.keys():
            resources = target_facts[fact]["resources"]
            self._actualize_resource_facts(resources, [])

        # Deleted fact category
        for fact in actual_facts.keys() - target_facts.keys():
            resources = actual_facts[fact]["resources"]
            self._actualize_resource_facts([], resources)

        # Actualize resource facts in the existing fact category
        for fact in target_facts.keys() & actual_facts.keys():
            target_resources = target_facts[fact]["resources"]
            actual_resources = actual_facts[fact]["resources"]
            self._actualize_resource_facts(target_resources, actual_resources)

    def _iteration(self):
        # The payload is collected every iteration from the data plane.
        # At the end of the iteration the payload hash is calculated
        # and it is saved
        collected_payload = models.Payload.empty()

        # Last successfully saved payload. Use it to compare with CP payload.
        if self._payload_path:
            last_payload = models.Payload.load(self._payload_path)
        else:
            last_payload = None

        # Check if the agent is registered
        try:
            payload = self._orch_client.agents_get_payload(
                self._agent_uuid, last_payload
            )
        except orch_exc.AgentNotFound:
            # Auto discovery mechanism
            self._register_agent()
            return

        # TODO(akremenetsky): Implement actions

        # Capabilities
        for capability, driver in self._caps_driver_map.items():
            # Skip capabilities that are not in the payload
            if capability not in payload.capabilities:
                LOG.debug("Skipping capability %s", capability)
                continue

            target_resources = payload.caps_resources(capability)

            try:
                collected_resources = self._actualize_resources(
                    driver, capability, target_resources
                )
                collected_payload.add_caps_resources(collected_resources)

                # All gathered resources for capabilities are considered
                # as facts too
                collected_payload.add_facts_resources(collected_resources)
            except Exception:
                LOG.exception("Error actualizing resources for %s", capability)

        # Facts
        for fact, driver in self._facts_driver_map.items():
            # Skip facts that are not in the payload
            if fact not in payload.facts:
                LOG.debug("Skipping fact %s", fact)
                continue

            try:
                collected_facts = driver.list(fact)
                collected_payload.add_facts_resources(collected_facts)
            except Exception:
                LOG.exception("Error collecting resources for fact: %s", fact)

        # All work done. The target resources are applied and facts collected.
        # Calculate the hash of the collected payload
        collected_payload.calculate_hash()

        # The payloads aren't the same. It means the facts were updated.
        if collected_payload != payload:
            self._actualize_facts(collected_payload.facts, payload.facts)

        # The capabilities were applied but the target and actual payloads
        # are different. So the difference in the facts. Update it.

        # Save the collected payload after actualization
        if self._payload_path:
            collected_payload.save(self._payload_path)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import os

from oslo_config import cfg

from restalchemy.storage.sql import engines
from restalchemy.storage.sql import migrations


CONF = cfg.CONF


INIT_MIGRATION_FILENAME = "0000-init-events-table-2cfd220e"


@contextlib.contextmanager
def _database_session(conf, db_type, name, **kwargs):
    conf = conf or CONF

    configure_database_factory = {
        "postgresql": engines.engine_factory.configure_postgresql_factory,
        "mysql": engines.engine_factory.configure_mysql_factory,
    }

    configure_database_factory[db_type](conf=conf, name=name, **kwargs)

    try:
        yield
    finally:
        engines.engine_factory.destroy_engine(name=name)


def _get_migration_engine():
    migration_engine = migrations.MigrationEngine(
        migrations_path=os.path.dirname(__file__)
    )
    return migration_engine


def apply_migrations(
    conf=None, db_type="postgresql", name=engines.DEFAULT_NAME, **kwargs
):
    with _database_session(conf, db_type, name=name, **kwargs):
        migration_engine = _get_migration_engine()
        last_migration = migration_engine.get_latest_migration()
        migration_engine.apply_migration(last_migration)


def rollback_migrations(
    conf=None, db_type="postgresql", name=engines.DEFAULT_NAME, **kwargs
):
    with _database_session(conf, db_type, name=name, **kwargs):
        migration_engine = _get_migration_engine()
        migration_engine.rollback_migration(INIT_MIGRATION_FILENAME)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import os
import json
import threading
from pathlib import Path


class JsonFileStorageSingleton(dict):
    _instances = {}
    _lock = threading.Lock()

    def __new__(cls, storage_path: str):
        if storage_path not in cls._instances:
            with cls._lock:
                if storage_path not in cls._instances:
                    cls._instances[storage_path] = super(
                        JsonFileStorageSingleton, cls
                    ).__new__(cls)
        return cls._instances[storage_path]

    def __init__(self, storage_path: str):
        self._storage_path = Path(storage_path)

        super().__init__()
        self.load()

    def load(self) -> None:
        """Load the storage from the file."""
        if not os.path.exists(self._storage_path):
            self.clear()
            return

        with open(self._storage_path) as f:
            data = json.load(f)

        self.clear()
        self.update(data)

    def persist(self) -> None:
        """Persist the storage to the file."""

        # Create the directory if it doesn't exist
        os.makedirs(os.path.dirname(self._storage_path), exist_ok=True)

        # Save the new data
        tmp_file = self._storage_path.with_suffix(".tmp")
        with open(tmp_file, "w") as f:
            json.dump(self, f, indent=2)
        os.replace(tmp_file, self._storage_path)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import os
import pwd
import grp
import logging
import subprocess

from restalchemy.dm import properties
from restalchemy.dm import types
from restalchemy.dm import types_dynamic

from gcl_sdk.agents.universal.drivers import meta
from gcl_sdk.agents.universal.drivers import exceptions as driver_exc
from gcl_sdk.agents.universal import constants as c


LOG = logging.getLogger(__name__)


class AbstractRenderHooks:
    def on_change(self) -> None:
        raise NotImplementedError()


class OnChangeNoAction(types_dynamic.AbstractKindModel, AbstractRenderHooks):
    KIND = "no_action"

    def on_change(self) -> None:
        # Do nothing
        pass


class OnChangeShell(types_dynamic.AbstractKindModel, AbstractRenderHooks):
    KIND = "shell"

    command = properties.property(
        types.String(max_length=262144), required=True, default=""
    )

    def on_change(self) -> None:
        subprocess.check_output(self.command, shell=True)


class Render(meta.MetaDataPlaneModel):
    """Render of the configuration file."""

    path = properties.property(
        types.String(min_length=1, max_length=512),
        required=True,
    )
    mode = properties.property(types.String(max_length=4), default="0644")
    owner = properties.property(
        types.String(max_length=128),
        default="root",
    )
    group = properties.property(
        types.String(max_length=128),
        default="root",
    )
    on_change = properties.property(
        types_dynamic.KindModelSelectorType(
            types_dynamic.KindModelType(OnChangeNoAction),
            types_dynamic.KindModelType(OnChangeShell),
        ),
    )
    content = properties.property(
        types.AllowNone(types.String()), default=None
    )

    def get_meta_model_fields(self) -> set[str] | None:
        """Return a list of meta fields or None.

        Meta fields are the fields that cannot be fetched from
        the data plane or we just want to save them into the meta file.

        `None` means all fields are meta fields but it doesn't mean they
        won't be updated from the data plane.
        """
        return {"uuid", "on_change", "path"}

    def dump_to_dp(self) -> None:
        """Save the render to the file system."""
        if self.content is None:
            raise ValueError("Render content is empty")

        # Create the directory if it doesn't exist
        if not os.path.exists(os.path.dirname(self.path)):
            os.makedirs(os.path.dirname(self.path))

        # Save the content
        with open(self.path, "w") as f:
            f.write(self.content)

        # Set the file mode, owner and group
        mode = int(self.mode, base=8)

        try:
            owner = pwd.getpwnam(self.owner).pw_uid
        except KeyError:
            raise ValueError(f"User {self.owner} does not exist")

        try:
            group = grp.getgrnam(self.group).gr_gid
        except KeyError:
            raise ValueError(f"Group {self.group} does not exist")

        os.chmod(self.path, mode)
        os.chown(self.path, owner, group)

        self.on_change.on_change()

        LOG.info("Render saved to %s", self.path)

    def restore_from_dp(self) -> None:
        """Load the render from the file system."""
        if not os.path.exists(self.path):
            resource = self.to_ua_resource("render")
            raise driver_exc.ResourceNotFound(resource=resource)

        with open(self.path) as f:
            self.content = f.read()

        # Read the file mode, owner and group
        stat = os.stat(self.path)
        self.mode = f"0{oct(stat.st_mode)[-3:]}"
        self.owner = pwd.getpwuid(stat.st_uid).pw_name
        self.group = grp.getgrgid(stat.st_gid).gr_name

    def delete_from_dp(self) -> None:
        """Delete the resource from the data plane."""
        if os.path.exists(self.path):
            os.remove(self.path)

    def update_on_dp(self) -> None:
        """Update the resource on the data plane."""
        # The simplest implementation, just recreate.
        self.delete_from_dp()
        self.dump_to_dp()


class RenderAgentDriver(meta.MetaFileStorageAgentDriver):
    RENDER_META_PATH = os.path.join(c.WORK_DIR, "render_meta.json")

    __model_map__ = {"render": Render}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, meta_file=self.RENDER_META_PATH, **kwargs)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import os
import pwd
import grp
import logging


from restalchemy.dm import properties
from restalchemy.dm import types

from gcl_sdk.agents.universal.drivers import meta
from gcl_sdk.agents.universal.drivers import exceptions
from gcl_sdk.agents.universal import constants as c


LOG = logging.getLogger(__name__)
AUTHORIZED_KEYS_PATH = ".ssh/authorized_keys"
SSH_KEY_TARGET_KIND = "ssh_key_target"


class SSHKey(meta.MetaDataPlaneModel):
    """SSH key model."""

    HOME = "/home"

    user = properties.property(types.String(min_length=1, max_length=64))
    authorized_keys = properties.property(
        types.String(min_length=1, max_length=256),
        default=AUTHORIZED_KEYS_PATH,
    )
    target_public_key = properties.property(
        types.String(max_length=10240),
        default="",
    )

    def get_meta_model_fields(self) -> set[str] | None:
        """Return a list of meta fields or None.

        Meta fields are the fields that cannot be fetched from
        the data plane or we just want to save them into the meta file.

        `None` means all fields are meta fields but it doesn't mean they
        won't be updated from the data plane.
        """

        # Keep all fields as meta fields for simplicity
        return {"uuid", "user", "authorized_keys", "target_public_key"}

    def dump_to_dp(self) -> None:
        """Save the key to the data plane."""

        # Check if authorized_keys file exists
        authorized_keys = os.path.join(
            self.HOME, self.user, self.authorized_keys
        )
        if not os.path.exists(authorized_keys):
            os.makedirs(os.path.dirname(authorized_keys), exist_ok=True)

            # Set the file mode, owner and group
            mode = int("0600", base=8)

            try:
                owner = pwd.getpwnam(self.user).pw_uid
            except KeyError:
                raise ValueError(f"User {self.user} does not exist")

            try:
                group = grp.getgrnam(self.user).gr_gid
            except KeyError:
                raise ValueError(f"Group {self.user} does not exist")

            with open(authorized_keys, "w") as f:
                f.write("")

            os.chmod(authorized_keys, mode)
            os.chown(authorized_keys, owner, group)

        # Add the key if it doesn't exist
        with open(authorized_keys, "r") as f:
            content = f.read()
        if self.target_public_key not in content:
            with open(authorized_keys, "a") as f:
                f.write(f"{self.target_public_key}\n")

    def restore_from_dp(self) -> None:
        """Load the key from the file system."""
        authorized_keys = os.path.join(
            self.HOME, self.user, self.authorized_keys
        )
        if not os.path.exists(authorized_keys):
            resource = self.to_ua_resource(SSH_KEY_TARGET_KIND)
            raise exceptions.ResourceNotFound(resource=resource)

        with open(authorized_keys, "r") as f:
            content = f.read()
        for key in content.split("\n"):
            if key == self.target_public_key:
                return

        resource = self.to_ua_resource(SSH_KEY_TARGET_KIND)
        raise exceptions.ResourceNotFound(resource=resource)

    def delete_from_dp(self) -> None:
        """Delete the key from the data plane."""
        authorized_keys = os.path.join(
            self.HOME, self.user, self.authorized_keys
        )
        if not os.path.exists(authorized_keys):
            return

        with open(authorized_keys, "r") as f:
            content = f.read()
        with open(authorized_keys, "w") as f:
            for key in content.split("\n"):
                if not key or key == "\n":
                    continue
                if key == self.target_public_key:
                    continue
                f.write(f"{key}\n")

    def update_on_dp(self) -> None:
        """Update the resource on the data plane."""
        # The simplest implementation, just recreate.
        self.delete_from_dp()
        self.dump_to_dp()


class SSHKeyCapabilityDriver(meta.MetaFileStorageAgentDriver):
    SSH_KEY_META_PATH = os.path.join(c.WORK_DIR, "ssh_key_meta.json")

    __model_map__ = {SSH_KEY_TARGET_KIND: SSHKey}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, meta_file=self.SSH_KEY_META_PATH, **kwargs)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import uuid as sys_uuid
import typing as tp
import logging
import contextlib

from restalchemy.common import contexts
from restalchemy.dm import filters as dm_filters
from restalchemy.storage import exceptions as ra_exc

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal.clients.orch import base
from gcl_sdk.agents.universal.clients.orch import exceptions


LOG = logging.getLogger(__name__)


class DatabaseOrchClient(base.AbstractOrchClient):
    """Database Orchestrator client.

    An interface for clients to communicate with an orchestrator.
    """

    @contextlib.contextmanager
    def _session_context(self, session: tp.Any | None = None) -> tp.Any:
        if session:
            yield session
        else:
            with contexts.Context().session_manager() as session:
                yield session

    def agents_create(
        self,
        agent: models.UniversalAgent,
        session: tp.Any = None,
    ) -> models.UniversalAgent:
        """Create an instance of Universal agent."""
        try:
            with self._session_context(session=session) as s:
                return agent.insert(session=s)
        except ra_exc.ConflictRecords:
            raise exceptions.AgentAlreadyExists(uuid=agent.uuid)

    def agents_get_payload(
        self,
        uuid: sys_uuid.UUID,
        payload: models.Payload | None,
        session: tp.Any = None,
    ) -> models.Payload:
        """Get payload for of the Universal agent."""
        if payload is None:
            payload = models.Payload.empty()

        with self._session_context(session=session):
            try:
                agent = models.UniversalAgent.objects.get_one(
                    filters={
                        "uuid": dm_filters.EQ(str(uuid)),
                    }
                )
            except ra_exc.RecordNotFound:
                raise exceptions.AgentNotFound(uuid=uuid)

            return agent.get_payload(
                hash=payload.hash, version=payload.version
            )

    def resources_create(
        self,
        resource: models.Resource,
        session: tp.Any = None,
    ) -> models.Resource:
        """Create a resource."""
        try:
            with self._session_context(session=session) as s:
                return resource.insert(session=s)
        except ra_exc.ConflictRecords:
            raise exceptions.ResourceAlreadyExists(uuid=resource.uuid)

    def resources_get(
        self, kind: str, uuid: sys_uuid.UUID, session: tp.Any = None
    ) -> models.Resource:
        """Get the resource."""
        try:
            with self._session_context(session=session):
                return models.Resource.objects.get_one(
                    filters={
                        "uuid": dm_filters.EQ(str(uuid)),
                        "kind": dm_filters.EQ(kind),
                    }
                )
        except ra_exc.RecordNotFound:
            raise exceptions.ResourceNotFound(uuid=uuid)

    def resources_update(
        self,
        kind: str,
        uuid: sys_uuid.UUID,
        session: tp.Any = None,
        **kwargs: tp.Any,
    ) -> models.Resource:
        """Update the resource."""
        with self._session_context(session=session) as s:
            resource = self.resources_get(kind, uuid, session=s)
            view = resource.dump_to_simple_view()
            view.update(**kwargs)
            resource = models.Resource.restore_from_simple_view(**view)
            return resource.update(session=s, force=True)

    def resources_delete(
        self, resource: models.Resource, session: tp.Any = None
    ) -> None:
        """Delete the resource."""
        with self._session_context(session=session) as s:
            resource = self.resources_get(
                resource.kind, resource.uuid, session=s
            )
            resource.delete(session=s)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import re
import typing as tp
import uuid as sys_uuid

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.clients.http import base as http
from gcl_sdk.agents.universal.clients.backend import rest
from gcl_sdk.agents.universal.clients.backend import exceptions


class ResourceProjectMismatch(exceptions.BackendClientException):
    __template__ = "The resource project mismatch: {resource}"
    resource: models.Resource


class GCRestApiBackendClient(rest.RestApiBackendClient):
    """Genesis Core Rest API backend client."""

    def __init__(
        self,
        http_client: http.CollectionBaseClient,
        collection_map: dict[str, str],
        project_id: sys_uuid.UUID,
    ) -> None:
        super().__init__(
            http_client=http_client, collection_map=collection_map
        )
        self._project_id = str(project_id)

    def create(self, resource: models.Resource) -> dict[str, tp.Any]:
        """Creates the resource. Returns the created resource."""
        # Inject mandatory fields
        resource.value["uuid"] = str(resource.uuid)

        # Simple validation for project_id. Only one project is supported.
        res_project_id = resource.value.get("project_id", None)
        if res_project_id and res_project_id != self._project_id:
            raise ResourceProjectMismatch(resource=resource)

        return super().create(resource)

    def update(self, resource: models.Resource) -> dict[str, tp.Any]:
        """Update the resource. Returns the updated resource."""
        # FIXME(akremenetsky): Not the best implementation
        # Remove popential RO fields
        value = resource.value.copy()
        resource.value.pop("created_at", None)
        resource.value.pop("updated_at", None)
        resource.value.pop("project_id", None)
        resource.value.pop("uuid", None)

        try:
            result = super().update(resource)
        finally:
            resource.value = value

        return result

    def list(self, kind: str) -> list[dict[str, tp.Any]]:
        """Lists all resources by kind."""
        return super().list(kind, project_id=self._project_id)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import functools
import uuid as sys_uuid
import typing as tp

import bazooka

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.clients.http import base


class UniversalAgentsClient(base.StaticCollectionBaseModelClient):
    __collection_path__ = "/v1/agents/"
    __model__ = models.UniversalAgent


class ResourcesClient(base.CollectionBaseModelClient):
    API_VERSION = "v1"

    __model__ = models.Resource

    def __init__(
        self,
        base_url: str,
        kind: str,
        http_client: bazooka.Client | None = None,
        auth: base.AbstractAuthenticator | None = None,
    ) -> None:
        collecton_path = f"/{self.API_VERSION}/kind/{kind}/resources/"
        super().__init__(base_url, collecton_path, http_client, auth)
        self._kind = kind

    def _set_kind_ref(self, resource: models.Resource | dict) -> None:
        """Set the resource kind as a reference.

        The `kind` field in the response is a reference since the kind is
        a collection now. It's not convenient to use it as the reference
        so set it to a reference before creating the resource in Status API.
        """
        ref = f"/{self.API_VERSION}/kind/{self._kind}"

        try:
            resource.kind = ref
        except AttributeError:
            if "kind" in resource:
                resource["kind"] = ref

    def _drop_kind_ref(self, resource: models.Resource | dict) -> None:
        """Drop the kind reference and set it as a string.

        The `kind` field in the response is a reference since the kind is
        a collection now. It's not convenient to use it as the reference
        so drop the reference and set the kind as a string before using it.
        """
        try:
            resource.kind = self._kind
        except AttributeError:
            if "kind" in resource:
                resource["kind"] = self._kind

    def get(self, uuid: sys_uuid.UUID) -> models.Resource:
        resource = super().get(uuid)
        self._drop_kind_ref(resource)
        return resource

    def filter(self, **filters: tp.Dict[str, tp.Any]) -> list[models.Resource]:
        resources = super().filter(**filters)
        for r in resources:
            self._drop_kind_ref(r)
        return resources

    def create(self, object: models.Resource) -> models.Resource:
        self._set_kind_ref(object)
        resource = super().create(object)
        self._drop_kind_ref(resource)
        return resource

    def update(
        self, uuid: sys_uuid.UUID, **params: tp.Dict[str, tp.Any]
    ) -> models.Resource:
        self._set_kind_ref(params)
        resource = super().update(uuid, **params)
        self._drop_kind_ref(resource)
        return resource


class StatusAPI:
    def __init__(
        self,
        base_url: str,
        http_client: bazooka.Client | None = None,
    ) -> None:
        http_client = http_client or bazooka.Client()
        self._http_client = http_client
        self._base_url = base_url
        self._agents_client = UniversalAgentsClient(base_url, http_client)

    @property
    def agents(self) -> UniversalAgentsClient:
        return self._agents_client

    @functools.lru_cache
    def resources(self, kind: str) -> ResourcesClient:
        return ResourcesClient(self._base_url, kind, self._http_client)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import logging
import uuid as sys_uuid

from gcl_sdk.agents.universal.drivers import base
from gcl_sdk.agents.universal.drivers import exceptions as driver_exc
from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal.clients.backend import base as client_base
from gcl_sdk.agents.universal.clients.backend import exceptions as client_exc
from gcl_sdk.agents.universal.storage import base as storage_base
from gcl_sdk.agents.universal.storage import exceptions as storage_exc

LOG = logging.getLogger(__name__)


class DirectAgentDriver(base.AbstractCapabilityDriver):
    """Direct driver. Directly gets all data from the backend.

    The key feature of this driver it's able to get all
    data from the data plane without using any additional
    information like meta files.

    It uses the target fields storage to store the target fields
    of the resources. This is needed to correctly calculate the
    hash of the resources.
    """

    def __init__(
        self,
        client: client_base.AbstractBackendClient,
        storage: storage_base.AbstractTargetFieldsStorage,
    ):
        super().__init__()
        self._client = client
        self._storage = storage

    def _model_to_resource(
        self,
        kind: str,
        model: models.ResourceMixin,
        target_fields: frozenset[str],
    ) -> models.Resource:
        # We need to be sure the return object is resource
        # and not target resource
        value = model.dump_to_simple_view(
            skip=model.get_resource_ignore_fields()
        )
        return models.Resource.from_value(value, kind, target_fields)

    def _validate(self, resource: models.Resource) -> None:
        """Validate the resource."""
        if resource.kind not in self.get_capabilities():
            raise TypeError(f"Unsupported capability {resource.kind}")

    def get(self, resource: models.Resource) -> models.Resource:
        """Find and return a resource by uuid and kind.

        It returns the resource from the backend.
        """
        self._validate(resource)

        try:
            target_fields = self._storage.get(resource.kind, resource.uuid)
            value = self._client.get(resource)
        except (client_exc.ResourceNotFound, storage_exc.ItemNotFound):
            LOG.error("Unable to find resource on backend %s", resource.uuid)
            raise driver_exc.ResourceNotFound(resource=resource)

        return resource.replace_value(value, target_fields.fields)

    def list(self, capability: str) -> list[models.Resource]:
        """Lists all resources by capability."""
        if capability not in self.get_capabilities():
            raise TypeError(f"Unsupported capability {capability}")

        # Collect all resources in convient format
        resources = []
        storage_items = {i.uuid: i for i in self._storage.list(capability)}
        client_resources = {}

        # Collect client items
        for i in self._client.list(capability):
            if isinstance(i, models.Resource):
                client_resources[i.uuid] = i
            elif isinstance(i, dict):
                uuid = sys_uuid.UUID(i["uuid"])
                if storage_item := storage_items.get(uuid):
                    client_resources[uuid] = models.Resource.from_value(
                        i, capability, storage_item.fields
                    )
                else:
                    LOG.warning("Missing storage item for %s", uuid)
            else:
                uuid = i.get_resource_uuid()

                # NOTE(akremenetsky): It's important point to take target fields
                # from storage but not from the models since fields can be
                # changed during live time and it will follow to wrong hash
                # and as a result migration problems. Seems it will be better
                # to explicitly update target fields during the update procedure.
                if storage_item := storage_items.get(uuid):
                    client_resources[uuid] = self._model_to_resource(
                        capability, i, storage_item.fields
                    )
                else:
                    LOG.warning("Missing storage item for %s", uuid)

        # If storage item or client item is missing, consider it as
        # a missing resource
        for uuid in storage_items.keys() & client_resources.keys():
            resources.append(client_resources[uuid])

        return resources

    def create(self, resource: models.Resource) -> models.Resource:
        """Creates a resource."""
        self._validate(resource)

        # Figure out the target fields to correct hash calculation
        target_fields = frozenset(resource.value.keys())
        storage_item = storage_base.TargetFieldItem(
            resource.kind, resource.uuid, target_fields
        )

        # There are no problems if the client fails its operation later.
        # The resource will be created next iteration.
        self._storage.create(storage_item, force=True)

        try:
            resp = self._client.create(resource)
            LOG.debug("Created resource: %s", resource.uuid)
        except client_exc.ResourceAlreadyExists:
            LOG.error("The resource already exists: %s", resource.uuid)
            raise driver_exc.ResourceAlreadyExists(resource=resource)

        # Convert response to the resource
        if isinstance(resp, models.Resource):
            return resp
        elif isinstance(resp, dict):
            return resource.replace_value(resp, target_fields)
        else:
            return self._model_to_resource(resource.kind, resp, target_fields)

    def update(self, resource: models.Resource) -> models.Resource:
        """Update the resource."""
        self._validate(resource)

        # Figure out the target fields to correct hash calculation
        target_fields = frozenset(resource.value.keys())
        storage_item = storage_base.TargetFieldItem(
            resource.kind, resource.uuid, target_fields
        )

        try:
            resp = self._client.update(resource)
            self._storage.update(storage_item)
            LOG.debug("Updated resource: %s", resource.uuid)
        except client_exc.ResourceNotFound:
            LOG.error("The resource does not exist: %s", resource.uuid)
            raise driver_exc.ResourceNotFound(resource=resource)

        # Convert response to the resource
        if isinstance(resp, models.Resource):
            return resp
        elif isinstance(resp, dict):
            return resource.replace_value(resp, target_fields)
        else:
            return self._model_to_resource(resource.kind, resp, target_fields)

    def delete(self, resource: models.Resource) -> None:
        """Delete the resource."""
        self._validate(resource)

        try:
            self._client.delete(resource)
            LOG.debug("Deleted resource: %s", resource.uuid)
        except client_exc.ResourceNotFound:
            LOG.warning("The resource is already deleted: %s", resource.uuid)

        storage_item = storage_base.TargetFieldItem(
            resource.kind, resource.uuid, frozenset()
        )

        self._storage.delete(storage_item, force=True)

    def start(self) -> None:
        """Perform some initialization before starting any operations.

        This method is called once before any other method like list,
        create, update, delete are called. It can be used to do some
        preparations like establishing connections, opening files, etc.

        The driver iteration:
            start -> list -> [create | update | delete]* -> finalize
        """
        self._storage.load()

    def finalize(self) -> None:
        """Perform some finalization after finishing all operations.

        This method is called once after all other methods like list,
        create, update, delete are called. It can be used to do some
        finalization or cleanups like closing connections, files, etc.

        The driver iteration:
            start -> list -> [create | update | delete]* -> finalize
        """
        self._storage.persist()



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import logging
import itertools
import typing as tp
import uuid as sys_uuid

from restalchemy.common import contexts
from restalchemy.dm import filters as dm_filters
from gcl_looper.services import basic as looper_basic

from gcl_sdk.common import constants as c

# from gcl_sdk.infra import constants as pc
from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal import constants as ua_c

LOG = logging.getLogger(__name__)


class UniversalBuilderService(looper_basic.BasicService):

    def __init__(
        self,
        instance_model: type[models.InstanceMixin],
        iter_min_period: float = 3,
        iter_pause: float = 0.1,
    ):
        super().__init__(iter_min_period, iter_pause)
        self._instance_model = instance_model

    # Builder interface

    def pre_create_instance_resource(
        self, instance: models.InstanceMixin
    ) -> None:
        """The hook is performed before creating instance resource.

        The hook is called only for new instances.
        """
        pass

    def create_instance_derivatives(
        self, instance: models.InstanceMixin
    ) -> tp.Collection[models.TargetResourceKindAwareMixin]:
        """Create the instance.

        The result is a collection of derivative objects that are
        required for the instance. For example, the main instance is a
        `Config` so the derivative objects for the the config is a list
        of `Render`. The result is a tuple/list/set/... of render objects.
        The derivative objects should inherit from the `TargetResourceMixin`.

        The hook is called only for new instances.
        """
        return tuple()

    def post_create_instance_resource(
        self,
        instance: models.InstanceMixin,
        resource: models.TargetResource,
        derivatives: tp.Collection[models.TargetResource] = tuple(),
    ) -> None:
        """The hook is performed after saving instance resource.

        The hook is called only for new instances.
        """
        instance.status = ua_c.InstanceStatus.IN_PROGRESS.value

    def pre_update_instance_resource(
        self, instance: models.InstanceMixin
    ) -> None:
        """The hook is performed before updating instance resource."""
        pass

    def update_instance_derivatives(
        self,
        instance: models.InstanceWithDerivativesMixin,
        resource: models.TargetResourceKindAwareMixin,
        derivative_pairs: tp.Collection[
            tuple[
                models.TargetResourceKindAwareMixin,  # The target resource
                models.TargetResourceKindAwareMixin
                | None,  # The actual resource
            ]
        ],
    ) -> tp.Collection[models.TargetResourceKindAwareMixin]:
        """The hook to update instance derivatives.

        The hook is called when an initiator of updating is an user or
        software from control plane side.
        The default behavior is to send the same list as on instance creation.
        """
        return self.create_instance_derivatives(instance)

    def post_update_instance_resource(
        self,
        instance: models.InstanceMixin,
        resource: models.TargetResource,
    ) -> None:
        """The hook is performed after updating instance resource."""
        pass

    def actualize_outdated_instance(
        self,
        current_instance: models.InstanceMixin,
        actual_instance: models.InstanceMixin,
    ) -> None:
        """Actualize outdated instance.

        It means some changes occurred on the data plane and the instance
        is outdated now. For example, the instance `Password` has field
        `value` that is stored in the secret storage. If the value is changed
        or created on the data plane, the instance is outdated and this method
        is called to reactualize the instance.

        Args:
            current_instance: The current instance.
            actual_instance: The actual instance.
        """
        pass

    def actualize_outdated_instance_derivatives(
        self,
        instance: models.InstanceMixin,
        derivative_pairs: tp.Collection[
            tuple[
                models.TargetResourceKindAwareMixin,  # The target resource
                models.TargetResourceKindAwareMixin
                | None,  # The actual resource
            ]
        ],
    ) -> tp.Collection[models.TargetResourceKindAwareMixin]:
        """Actualize outdated instance with derivatives.

        It means some changes occurred on the data plane and the instance
        is outdated now. For example, the instance `Config` has derivative
        `Render`. Single `Config` may have multiple `Render` derivatives.
        If any of the derivatives is outdated, this method is called to
        reactualize the derivatives. The method returns the list of `updated`
        derivatives. If nothing needs to be updated, the method returns the
        same list of target derivatives as it received. Otherwise, the method
        should return the list of updated derivatives. It also can add new or
        remove old derivatives.
        Depends on the `fetch_all_derivatives_on_outdate` the behavior of the
        method is different:

        fetch_all_derivatives_on_outdate == True:
        The method receives the list of all derivatives currently available
        for the instance even though the derivatives aren't outdated.

        fetch_all_derivatives_on_outdate == False:
        The method receives the list only changed derivatives from the last
        actualization. For example, a config has two renders. Only one of
        them is outdated. The method receives the list of only one outdated
        render.

        Args:
            instance: The instance to actualize.
            derivative_pairs: Changed or all derivatives of the instance.
        """
        return tuple(p[0] for p in derivative_pairs)

    def pre_delete_instance_resource(
        self, resource: models.TargetResource
    ) -> None:
        """The hook is performed before deleting instance resource."""
        pass

    def track_outdated_master_hash_instances(self) -> bool:
        """Track outdated master hash instances."""
        return False

    def track_outdated_master_full_hash_instances(self) -> bool:
        """Track outdated master full hash instances."""
        return False

    def actualize_outdated_master_hash_instance(
        self,
        instance: models.InstanceMixin,
        master_instance: models.InstanceMixin,
        derivatives: tp.Collection[
            tuple[
                models.TargetResourceKindAwareMixin,  # The target resource
                models.TargetResourceKindAwareMixin
                | None,  # The actual resource
            ]
        ],
    ) -> tp.Collection[models.TargetResourceKindAwareMixin]:
        """Actualize outdated master hash instance.

        The logic is quite similar to `actualize_outdated_instance_derivatives`.
        But the reason when this method is called is different. The
        `actualize_outdated_instance_derivatives` allows to track changes on the
        data plane but this method allows to track changes on a related master
        instance. For example, the instance model is `Database`, the related master
        for this instance is `NodeSet`. If the `NodeSet` is updated, this method
        is called for all `Database` instances that are related to this `NodeSet`
        to reactualize them.
        This method tracks changes for target fields of the master instance.

        Args:
            instance: The instance to actualize.
            master_instance: The master instance.
            derivatives: All derivatives of the instance.
        """
        return tuple(p[0] for p in derivatives)

    def actualize_outdated_master_full_hash_instance(
        self,
        instance: models.InstanceMixin,
        master_instance: models.InstanceMixin,
        derivatives: tp.Collection[
            tuple[
                models.TargetResourceKindAwareMixin,  # The target resource
                models.TargetResourceKindAwareMixin
                | None,  # The actual resource
            ]
        ],
    ) -> tp.Collection[models.TargetResourceKindAwareMixin]:
        """Actualize outdated master full hash instance.

        The logic is quite similar to `actualize_outdated_instance_derivatives`.
        But the reason when this method is called is different. The
        `actualize_outdated_instance_derivatives` allows to track changes on the
        data plane but this method allows to track changes on a related master
        instance. For example, the instance model is `Database`, the related master
        for this instance is `NodeSet`. If the `NodeSet` is updated, this method
        is called for all `Database` instances that are related to this `NodeSet`
        to reactualize them.
        This method tracks changes for all fields of the master instance.

        Args:
            instance: The instance to actualize.
            master_instance: The master instance.
            derivatives: All derivatives of the instance.
        """
        return tuple(p[0] for p in derivatives)

    # Internal methods

    def _are_target_resources_equal(
        self,
        resources_left: tp.Collection[models.TargetResource],
        resources_right: tp.Collection[models.TargetResource],
    ) -> bool:
        """Compare two collections of target resources."""
        left_hash = hash(frozenset(r.hash for r in resources_left))
        right_hash = hash(frozenset(r.hash for r in resources_right))
        return left_hash == right_hash

    def _actualize_resource_hash_status(
        self,
        target_resource: models.TargetResource,
        actual_resource: models.Resource,
        active_status: str = ua_c.InstanceStatus.ACTIVE.value,
    ) -> None:
        """Actualize resource hash and status."""
        target_resource.full_hash = actual_resource.full_hash

        # `ACTIVE` only if the hash is the same
        if (
            actual_resource.status == active_status
            and target_resource.hash == actual_resource.hash
        ):
            target_resource.status = actual_resource.status
        elif (
            actual_resource.status != active_status
            and target_resource.status != actual_resource.status
        ):
            target_resource.status = actual_resource.status
        target_resource.update()
        LOG.debug("Outdated resource %s actualized", target_resource.uuid)

    def _actualize_derivative_resources_hash_status(
        self,
        derivatives: tp.Collection[
            tuple[models.TargetResource, models.Resource]
        ],
        active_status: str = ua_c.InstanceStatus.ACTIVE.value,
    ) -> None:
        """Actualize derivative resources hash and status."""
        # Update target derivatives with actual information from the DP.
        for target_derivative, actual_derivative in derivatives:
            self._actualize_resource_hash_status(
                target_derivative,
                actual_derivative,
                active_status,
            )

    def _actualize_derivative_target_resources(
        self,
        new_target_resources: frozenset[models.TargetResource],
        current_target_resources: frozenset[models.TargetResource],
    ) -> None:
        """Actualize derivative target resources.

        A new collection of target resources is generated by some reason.
        This method compares the new collection with the current collection
        and actualizes the target resources.
        """
        # Compare if the resources are the same
        if self._are_target_resources_equal(
            new_target_resources, current_target_resources
        ):
            return

        # Create new target resources
        for resource in new_target_resources - current_target_resources:
            resource.insert()
            LOG.debug(
                "New target resource(%s) %s created",
                resource.kind,
                resource.uuid,
            )

        # Delete outdated target resources
        for resource in current_target_resources - new_target_resources:
            resource.delete()
            LOG.debug(
                "Outdated target resource(%s) %s deleted",
                resource.kind,
                resource.uuid,
            )

        # Update target resources
        # FIXME(akremenetsky): Both collections are expected to be small.
        # So we can use nested loop here.
        for new_resource in new_target_resources:
            for current_resource in current_target_resources:
                # The resource has been changed
                # We need to recreate it.
                if (
                    new_resource == current_resource
                    and new_resource.hash != current_resource.hash
                ):
                    current_resource.update_value(new_resource)
                    current_resource.update()
                    LOG.debug(
                        "Target resource(%s) %s recreated",
                        new_resource.kind,
                        new_resource.uuid,
                    )

    # `Outdated` actualization methods

    def _actualize_outdated_instance(
        self,
        instance: models.InstanceMixin,
        target_resource: models.TargetResource,
        actual_resource: models.Resource,
        active_status: str = ua_c.InstanceStatus.ACTIVE.value,
    ) -> None:
        """Actualize outdated instance.

        It means some changes occurred on the data plane and the instance
        is outdated now. For example, its status is incorrect.
        Usually, this method has to update status and full_hash of the instance.

        Args:
            instance: The instance to actualize.
            target_resource: The target resource of the instance.
            actual_resource: The actual resource of the instance.
        """
        actual_instance = self._instance_model.from_ua_resource(
            actual_resource
        )

        # Hook to actualize instance. For example, for `Password` instance
        # we need to update its value, saved into the secret storage.
        self.actualize_outdated_instance(instance, actual_instance)

        self._actualize_resource_hash_status(
            target_resource,
            actual_resource,
            active_status,
        )

        # Save the instance if some changes occurred.
        instance.save()

        # Commit the tracked_at timestamp
        target_resource.tracked_at = instance.updated_at
        target_resource.status = actual_resource.status
        target_resource.update()

        LOG.info(
            "Instance(%s) resource %s actualized",
            instance.get_resource_kind(),
            target_resource.uuid,
        )

    def _actualize_outdated_instance_derivatives(
        self,
        instance: models.InstanceWithDerivativesMixin,
        target_resource: models.TargetResource,
        actual_resource: models.Resource | None = None,
        changed_derivatives: tp.Collection[
            tuple[models.TargetResource, models.Resource]
        ] = tuple(),
        active_status: str | None = ua_c.InstanceStatus.ACTIVE.value,
    ) -> None:
        """Actualize outdated instance.

        It means some changes occurred on the data plane and the instance
        is outdated now. For example, its status is incorrect. Usually, this
        method has to update status and full_hash of the instance and
        derivatives. This method accepts only changed derivatives.
        So if you need to fetch all derivatives both changed and not changed
        use `actualize_outdated_instance_all_derivatives`.

        Args:
            instance: The instance to actualize.
            target_resource: The target resource of the instance.
            actual_resource: The actual resource of the instance.
            changed_derivatives: The changed derivatives of the instance.
        """
        if len(changed_derivatives) == 0:
            return

        # Convert derivatives to actual models
        derivatives = [
            (
                self._instance_model.derivative_model(t.kind).from_ua_resource(
                    t
                ),
                self._instance_model.derivative_model(a.kind).from_ua_resource(
                    a
                ),
            )
            for t, a in changed_derivatives
        ]

        # Hook to actualize instance.
        # Pay attention. This derivatives collection is only a subset of the
        # derivatives of the instance. It means that some derivatives may
        # be not changed. So all operation below are applied to the subset
        # of derivatives.
        new_derivatives = self.actualize_outdated_instance_derivatives(
            instance, derivatives
        )

        # Actualize hash and status of the target resources
        if actual_resource is not None:
            self._actualize_resource_hash_status(
                target_resource,
                actual_resource,
                active_status,
            )
        self._actualize_derivative_resources_hash_status(changed_derivatives)

        new_target_resources = frozenset(
            d.to_ua_resource(master=target_resource.uuid)
            for d in new_derivatives
        )
        # Actualize derivative target resources if they are changed
        self._actualize_derivative_target_resources(
            new_target_resources, frozenset(d for d, _ in changed_derivatives)
        )

        # Commit the tracked_at timestamp
        instance.update()
        target_resource.tracked_at = instance.updated_at
        target_resource.status = instance.status
        if actual_resource is not None:
            target_resource.full_hash = actual_resource.full_hash
        target_resource.update()

    def _actualize_outdated_instance_all_derivatives(
        self,
        instance: models.InstanceWithDerivativesMixin,
        target_resource: models.TargetResource,
        actual_resource: models.Resource | None = None,
        changed_derivatives: tp.Collection[
            tuple[models.TargetResource, models.Resource]
        ] = tuple(),
        all_derivatives: tp.Collection[
            tuple[models.TargetResource, models.Resource | None]
        ] = tuple(),
        active_status: str | None = ua_c.InstanceStatus.ACTIVE.value,
    ) -> None:
        """Actualize outdated instance.

        It means some changes occurred on the data plane and the instance
        is outdated now. For example, its status is incorrect. Usually, this
        method has to update status and full_hash of the instance and
        derivatives. This method accepts all derivatives both changed and
        not changed. So if you need to fetch only changed derivatives use
        `actualize_outdated_instance_derivatives`.

        Args:
            instance: The instance to actualize.
            target_resource: The target resource of the instance.
            actual_resource: The actual resource of the instance.
            changed_derivatives: The changed derivatives of the instance.
            all_derivatives: The all derivatives of the instance.
        """
        if len(changed_derivatives) == 0:
            return

        # Convert derivatives to actual models
        derivatives = [
            (
                self._instance_model.derivative_model(t.kind).from_ua_resource(
                    t
                ),
                (
                    self._instance_model.derivative_model(
                        a.kind
                    ).from_ua_resource(a)
                    if a is not None
                    else None
                ),
            )
            for t, a in all_derivatives
        ]

        # Hook to actualize instance.
        # Look at the docstring of the method for more details.
        new_derivatives = self.actualize_outdated_instance_derivatives(
            instance, derivatives
        )

        # Actualize hash and status of the target resources
        if actual_resource is not None:
            self._actualize_resource_hash_status(
                target_resource,
                actual_resource,
                active_status,
            )
        self._actualize_derivative_resources_hash_status(changed_derivatives)

        new_target_resources = frozenset(
            d.to_ua_resource(master=target_resource.uuid)
            for d in new_derivatives
        )

        # Actualize derivative target resources if they are changed
        self._actualize_derivative_target_resources(
            new_target_resources, frozenset(d for d, _ in all_derivatives)
        )

        # Commit the tracked_at timestamp
        instance.update()
        target_resource.tracked_at = instance.updated_at
        target_resource.status = getattr(
            instance, "status", ua_c.InstanceStatus.ACTIVE.value
        )
        if actual_resource is not None:
            target_resource.full_hash = actual_resource.full_hash
        target_resource.update()

    def _actualize_new_instance(self, instance: models.InstanceMixin) -> None:
        """Actualize the new PaaS instance."""
        # Perform some additional actions before creating
        # resource and derivatives
        self.pre_create_instance_resource(instance)

        # Create nessary derivatives for the instance
        if self._instance_model._has_model_derivatives():
            derivative_objects = self.create_instance_derivatives(instance)
        else:
            derivative_objects = tuple()

        # Convert instance to resource
        instance_resource = instance.to_ua_resource()
        instance_resource.insert()

        # Save the derivative objects
        derivative_resources = []
        for derivative_object in derivative_objects:
            derivative_resource = derivative_object.to_ua_resource(
                master=instance_resource.uuid
            )
            derivative_resources.append(derivative_resource)
            derivative_resource.save()

        self.post_create_instance_resource(
            instance, instance_resource, derivative_resources
        )
        for derivative_resource in derivative_resources:
            derivative_resource.update()

        instance.save()

        # Commit tracked_at ts
        instance_resource.tracked_at = instance.updated_at
        instance_resource.status = instance.status

        instance_resource.update()

        LOG.info("Instance resource %s created", instance_resource.uuid)

    def _actualize_new_instances(
        self, instances: tp.Collection[models.InstanceMixin] = tuple()
    ) -> None:
        """Actualize new PaaS instances."""
        instances = instances or self._instance_model.get_new_instances()

        if len(instances) == 0:
            return

        # Create resources for new instances
        for instance in instances:
            try:
                self._actualize_new_instance(instance)
            except Exception:
                LOG.exception(
                    "Error creating instance resource %s", instance.uuid
                )

    # `Updated` actualization methods

    def _actualize_updated_instance(
        self,
        instance: models.InstanceMixin,
        resource: models.TargetResource,
        derivatives: tp.Collection[
            tuple[models.TargetResource, models.Resource | None]
        ] = tuple(),
    ) -> None:
        """Actualize updated instance by user."""
        # Perform some additional actions before updating
        # resource and derivatives
        self.pre_update_instance_resource(instance)

        new_resource = instance.to_ua_resource()

        # Update the original resource
        resource.update_value(new_resource)

        # NOTE(akremenetsky): The default implementation of the method
        # `update_instance_derivatives` may be dangerous in cases if the
        # derivatives are changed in live cycle of the instance. For instance,
        # if new derivatives are added after the instance was created, they
        # will be dropped in the default implementation of the method.
        if self._instance_model._has_model_derivatives():
            # Convert derivative resources to actual models
            current_derivatives = tuple(
                (
                    self._instance_model.derivative_model(
                        t.kind
                    ).from_ua_resource(t),
                    (
                        self._instance_model.derivative_model(
                            a.kind
                        ).from_ua_resource(a)
                        if a is not None
                        else None
                    ),
                )
                for t, a in derivatives
            )

            # Create nessary derivatives for the updated instance
            new_derivatives = self.update_instance_derivatives(
                instance, resource, current_derivatives
            )
            new_target_resources = frozenset(
                d.to_ua_resource(master=resource.uuid) for d in new_derivatives
            )

            # Actualize derivative target resources if they are changed
            self._actualize_derivative_target_resources(
                new_target_resources, frozenset(d for d, _ in derivatives)
            )

        instance.save()

        # Commit the tracked_at timestamp
        resource.tracked_at = instance.updated_at
        resource.status = instance.status
        resource.update()

        self.post_update_instance_resource(instance, resource)

        LOG.info(
            "Instance(%s) resource %s updated",
            instance.get_resource_kind(),
            resource.uuid,
        )

    def _actualize_updated_instances(self) -> None:
        """Actualize updated instances changed by user."""
        updated_instances = self._instance_model.get_updated_instances()

        if len(updated_instances) == 0:
            return

        instance_resources = models.TargetResource.objects.get_all(
            filters={
                "uuid": dm_filters.In(str(i.uuid) for i in updated_instances),
                "kind": dm_filters.EQ(
                    self._instance_model.get_resource_kind()
                ),
            }
        )

        # Derivatives resources if the model class has derivatives
        if updated_instances[0]._has_model_derivatives():
            resource_derivative_map = self._get_resources_by_masters(
                tuple(r.uuid for r in instance_resources)
            )
        else:
            resource_derivative_map = {}

        resource_map = {i.uuid: i for i in instance_resources}

        # Update every resource in accordance with the new secret
        for instance in updated_instances:
            resource = resource_map[instance.uuid]
            derivatives = resource_derivative_map.get(resource.uuid, tuple())

            try:
                self._actualize_updated_instance(
                    instance, resource, derivatives
                )
            except Exception:
                LOG.exception(
                    "Error updating instance(%s) resource %s",
                    instance.get_resource_kind(),
                    resource.uuid,
                )

    def _get_outdated_instance_resources(
        self,
        limit: int = c.DEF_SQL_LIMIT,
    ) -> dict[
        sys_uuid.UUID,  # Resource UUID
        tuple[models.TargetResource, models.Resource],
    ]:
        kind = self._instance_model.get_resource_kind()
        outdated = models.OutdatedResource.objects.get_all(
            filters={"kind": dm_filters.EQ(kind)},
            limit=limit,
        )
        return {
            pair.target_resource.uuid: (
                pair.target_resource,
                pair.actual_resource,
            )
            for pair in outdated
        }

    def _actualize_outdated_instances_no_derivatives(self) -> None:
        """Actualize outdated instances.

        It means some changes occurred on the data plane and the instances
        are outdated now. For instance, their status is incorrect.

        This method is only used when the model class doesn't have
        derivatives.
        """
        kind = self._instance_model.get_resource_kind()

        # Check if there are outdated instance resources
        resource_map = self._get_outdated_instance_resources()

        if len(resource_map) == 0:
            return

        instances = self._instance_model.objects.get_all(
            filters={
                "uuid": dm_filters.In(str(u) for u in resource_map.keys()),
            },
        )

        # Actualize outdated instances. At least need to check
        # the status and full_hash
        for instance in instances:
            target, actual = resource_map[instance.uuid]
            try:
                self._actualize_outdated_instance(instance, target, actual)
                LOG.debug("Instance(%s) %s actualized", kind, instance.uuid)
            except Exception:
                LOG.exception(
                    "Error actualizing instance(%s) %s", kind, instance.uuid
                )

    def _get_outdated_derivative_resources(
        self,
        limit: int = c.DEF_SQL_LIMIT,
    ) -> dict[
        sys_uuid.UUID,  # Master UUID
        list[tuple[models.TargetResource, models.Resource]],
    ]:
        outdated = models.OutdatedResource.objects.get_all(
            filters={
                "kind": dm_filters.In(self._instance_model.derivative_kinds())
            },
            limit=limit,
        )
        key_map = {}
        for pair in outdated:
            key_map.setdefault(pair.target_resource.master, []).append(
                (pair.target_resource, pair.actual_resource)
            )

        return key_map

    def _get_outdated_instances_by_masters(
        self,
        masters: tp.Collection[sys_uuid.UUID],
    ) -> list[tuple[models.InstanceMixin, models.TargetResource]]:
        kind = self._instance_model.get_resource_kind()
        resources = models.TargetResource.objects.get_all(
            filters={
                "uuid": dm_filters.In(m for m in masters),
                "kind": dm_filters.EQ(kind),
            },
            order_by={"uuid": "asc"},
        )

        instances = self._instance_model.objects.get_all(
            filters={
                "uuid": dm_filters.In(m for m in masters),
            },
            order_by={"uuid": "asc"},
        )

        if len(instances) != len(resources):
            raise RuntimeError("Number of instances and resources not equal")

        return list(zip(instances, resources))

    def _get_resources_by_masters(
        self,
        masters: tp.Collection[sys_uuid.UUID],
    ) -> dict[
        sys_uuid.UUID,  # Master UUID
        list[tuple[models.TargetResource, models.Resource | None]],
    ]:
        kinds = self._instance_model.derivative_kinds()
        target_resources = models.TargetResource.objects.get_all(
            filters={
                "master": dm_filters.In(m for m in masters),
                "kind": dm_filters.In(kinds),
            },
            order_by={"uuid": "asc"},
        )

        # Prepare actual resources map
        actual_resources = {
            (r.uuid, r.kind): r
            for r in models.Resource.objects.get_all(
                filters={
                    "uuid": dm_filters.In(r.uuid for r in target_resources),
                    "kind": dm_filters.In(kinds),
                },
                order_by={"uuid": "asc"},
            )
        }

        # Prepare pairs of target and actual resources
        resource_map = {}
        for target_resource in target_resources:
            actual_resource = actual_resources.get(
                (target_resource.uuid, target_resource.kind)
            )
            resource_map.setdefault(target_resource.master, []).append(
                (target_resource, actual_resource)
            )

        return resource_map

    def _actualize_outdated_instances_have_derivatives(self) -> None:
        """Actualize outdated instances.

        It means some changes occurred on the data plane and the instances
        are outdated now. For instance, their status is incorrect.

        This method is only used when the model class has derivatives.
        For example, `Config` has derivatives `Render`.
        """
        changed_resource_map = self._get_outdated_derivative_resources()

        # Nothing to actualize if no changed resources
        if len(changed_resource_map) == 0:
            return

        need_all = self._instance_model.fetch_all_derivatives_on_outdate()

        # Get all instances related to changed resources
        instances = self._get_outdated_instances_by_masters(
            tuple(changed_resource_map.keys())
        )

        # Also fetch all derivatives if needed
        if need_all:
            all_resource_map = self._get_resources_by_masters(
                tuple(changed_resource_map.keys())
            )

        for instance, resource in instances:
            changed_derivatives = changed_resource_map[resource.uuid]

            try:
                # Actualize instance with all derivatives
                if need_all:
                    all_derivatives = all_resource_map[resource.uuid]
                    self._actualize_outdated_instance_all_derivatives(
                        instance,
                        resource,
                        changed_derivatives=changed_derivatives,
                        all_derivatives=all_derivatives,
                    )
                else:
                    # Actualize instance with only changed derivatives
                    self._actualize_outdated_instance_derivatives(
                        instance,
                        resource,
                        changed_derivatives=changed_derivatives,
                    )
            except Exception:
                LOG.exception(
                    "Error actualizing outdated instance %s", instance.uuid
                )

    def _actualize_outdated_instances(self) -> None:
        """Actualize outdated instances.

        It means some changes occurred on the data plane and the instances
        are outdated now. For instance, their status is incorrect.
        """
        if self._instance_model._has_model_derivatives():
            self._actualize_outdated_instances_have_derivatives()
        else:
            self._actualize_outdated_instances_no_derivatives()

    # `Deleted` actualization methods

    def _actualize_deleted_instances(self) -> None:
        """Actualize deleted instances."""
        deleted_instance_resources = (
            self._instance_model.get_deleted_instances()
        )

        if len(deleted_instance_resources) == 0:
            return

        if self._instance_model._has_model_derivatives():
            derivative_resources = models.TargetResource.objects.get_all(
                filters={
                    "master": dm_filters.In(
                        str(m.uuid) for m in deleted_instance_resources
                    ),
                }
            )
        else:
            derivative_resources = tuple()

        # Delete all outdated resources
        for resource in itertools.chain(
            deleted_instance_resources, derivative_resources
        ):
            try:
                self.pre_delete_instance_resource(resource)
                resource.delete()
                LOG.info(
                    "Outdated resource(%s) %s deleted",
                    resource.kind,
                    resource.uuid,
                )
            except Exception:
                LOG.exception("Error deleting resource %s", resource.uuid)

    # Outdated master hash (full hash) instances

    def _get_outdated_master_hash_resources(
        self,
        limit: int = c.DEF_SQL_LIMIT,
    ) -> tuple[tuple[models.TargetResource, models.TargetResource], ...]:
        kind = self._instance_model.get_resource_kind()
        outdated = models.OutdatedMasterHashResource.objects.get_all(
            filters={"kind": dm_filters.EQ(kind)},
            limit=limit,
        )
        return tuple(
            (
                pair.target_resource,
                pair.master,
            )
            for pair in outdated
        )

    def _get_outdated_master_full_hash_resources(
        self,
        limit: int = c.DEF_SQL_LIMIT,
    ) -> tuple[tuple[models.TargetResource, models.TargetResource], ...]:
        kind = self._instance_model.get_resource_kind()
        outdated = models.OutdatedMasterFullHashResource.objects.get_all(
            filters={"kind": dm_filters.EQ(kind)},
            limit=limit,
        )
        return tuple(
            (
                pair.target_resource,
                pair.master,
            )
            for pair in outdated
        )

    def _actualize_outdated_master_hash_instance(
        self,
        target_resource: models.TargetResource,
        master: models.TargetResource,
        derivatives: tp.Collection[
            tuple[models.TargetResource, models.Resource | None]
        ],
        tracked_field: tp.Literal["hash", "full_hash"] = "hash",
    ) -> None:
        """Actualize outdated master hash instance."""
        instance = self._instance_model.from_ua_resource(target_resource)

        if self._instance_model.__master_model__ is None:
            raise ValueError("The master model is not initialized.")

        master_instance = (
            self._instance_model.__master_model__.from_ua_resource(master)
        )

        converted_derivatives = [
            (
                self._instance_model.derivative_model(t.kind).from_ua_resource(
                    t
                ),
                (
                    self._instance_model.derivative_model(
                        a.kind
                    ).from_ua_resource(a)
                    if a is not None
                    else None
                ),
            )
            for t, a in derivatives
        ]

        # Hook to actualize instance.
        # Look at the docstring of the method for more details.
        if tracked_field == "hash":
            new_derivatives = self.actualize_outdated_master_hash_instance(
                instance, master_instance, converted_derivatives
            )
        else:
            new_derivatives = (
                self.actualize_outdated_master_full_hash_instance(
                    instance, master_instance, converted_derivatives
                )
            )

        new_target_resources = frozenset(
            d.to_ua_resource(master=target_resource.uuid)
            for d in new_derivatives
        )

        # Actualize derivative target resources if they are changed
        self._actualize_derivative_target_resources(
            new_target_resources, frozenset(d for d, _ in derivatives)
        )

        # Commit the master hash
        if tracked_field == "hash":
            target_resource.master_hash = master.hash
        else:
            target_resource.master_full_hash = master.full_hash
        target_resource.save()

    def _actualize_outdated_master_instances(
        self,
        resources: tuple[
            tuple[models.TargetResource, models.TargetResource], ...
        ],
        tracked_field: tp.Literal["hash", "full_hash"] = "hash",
    ) -> None:
        """Actualize outdated master hash instances.

        Track changes for the related master instances.
        Look at `actualize_outdated_master_hash_instance` for more details.
        """
        if len(resources) == 0:
            return

        # Get all derivatives related to the instances.
        # We need to be clear here. Masters for this call are instances
        # (resources) and not masters of the instances.
        # For example, our instance model is `Database`. The master of the
        # instance is `NodeSet` but for the call we pass `Database` as
        # master to get their derivatives and not derivatives of the
        # `NodeSet`.
        derivative_map = self._get_resources_by_masters(
            tuple(r.uuid for r, _ in resources)
        )

        for target_resource, master in resources:
            derivatives = derivative_map[target_resource.uuid]
            try:
                self._actualize_outdated_master_hash_instance(
                    target_resource,
                    master,
                    derivatives=derivatives,
                    tracked_field=tracked_field,
                )
            except Exception:
                LOG.exception(
                    "Error actualizing outdated master hash instance %s",
                    target_resource.uuid,
                )

    def _actualize_outdated_master_hash_instances(self) -> None:
        """Actualize outdated master hash instances."""

        resources = self._get_outdated_master_hash_resources()
        self._actualize_outdated_master_instances(
            resources, tracked_field="hash"
        )

    def _actualize_outdated_master_full_hash_instances(self) -> None:
        """Actualize outdated master full hash instances."""

        resources = self._get_outdated_master_full_hash_resources()
        self._actualize_outdated_master_instances(
            resources, tracked_field="full_hash"
        )

    # Misc methods

    def _iteration(self) -> None:
        with contexts.Context().session_manager():
            try:
                self._actualize_new_instances()
            except Exception:
                LOG.exception("Error actualizing new instances")

            try:
                self._actualize_deleted_instances()
            except Exception:
                LOG.exception("Error actualizing deleted instances")

            try:
                self._actualize_updated_instances()
            except Exception:
                LOG.exception("Error actualizing updated instances")

            try:
                self._actualize_outdated_instances()
            except Exception:
                LOG.exception("Error actualizing outdated instances")

            if self.track_outdated_master_hash_instances():
                try:
                    self._actualize_outdated_master_hash_instances()
                except Exception:
                    LOG.exception(
                        "Error actualizing outdated master hash instances"
                    )

            if self.track_outdated_master_full_hash_instances():
                try:
                    self._actualize_outdated_master_full_hash_instances()
                except Exception:
                    LOG.exception(
                        "Error actualizing outdated master full hash instances"
                    )



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

from urllib.parse import urljoin
import uuid as sys_uuid

import pytest
import requests
from oslo_config import cfg

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.tests.functional import utils as test_utils
from gcl_sdk.tests.functional import conftest


CONF = cfg.CONF


class TestUAStatusApi:
    @pytest.fixture(scope="class")
    def status_api_service(self, status_api_wsgi_app):
        class ApiRestService(test_utils.RestServiceTestCase):
            __FIRST_MIGRATION__ = conftest.FIRST_MIGRATION
            __APP__ = status_api_wsgi_app

        rest_service = ApiRestService()
        rest_service.setup_class()

        yield rest_service

        rest_service.teardown_class()

    @pytest.fixture()
    def status_api(self, status_api_service: test_utils.RestServiceTestCase):
        status_api_service.setup_method()

        yield status_api_service

        status_api_service.teardown_method()

    def test_agent_no_agents(
        self,
        status_api: test_utils.RestServiceTestCase,
    ):
        url = urljoin(status_api.base_url, "agents/")

        response = requests.get(url)

        assert response.status_code == 200

    def test_agent_list(
        self,
        status_api: test_utils.RestServiceTestCase,
    ):
        uuid_a = sys_uuid.uuid4()
        agent_a = models.UniversalAgent(
            name="Agent A", uuid=uuid_a, node=uuid_a
        )
        uuid_b = sys_uuid.uuid4()
        agent_b = models.UniversalAgent(
            name="Agent B", uuid=uuid_b, node=uuid_b
        )
        agent_a.insert()
        agent_b.insert()

        url = urljoin(status_api.base_url, "agents/")

        response = requests.get(url)
        output = response.json()

        assert response.status_code == 200
        assert len(output) == 2
        assert {a["uuid"] for a in output} == {str(uuid_a), str(uuid_b)}

    def test_agent_register(
        self,
        status_api: test_utils.RestServiceTestCase,
    ):
        uuid = sys_uuid.uuid4()
        agent = models.UniversalAgent(
            name="Agent A",
            uuid=uuid,
            node=uuid,
            capabilities={"capabilities": ["foo"]},
            facts={"facts": ["bar"]},
        )

        view = agent.dump_to_simple_view()

        url = urljoin(status_api.base_url, "agents/")

        response = requests.post(url, json=view)
        output = response.json()

        assert response.status_code == 201
        assert output["uuid"] == str(uuid)

    def test_resources_no_kind(
        self,
        status_api: test_utils.RestServiceTestCase,
    ):
        url = urljoin(status_api.base_url, "kind/")

        response = requests.get(url)
        output = response.json()

        assert response.status_code == 200
        assert output == []

    def test_resources_empty_kind(
        self,
        status_api: test_utils.RestServiceTestCase,
    ):
        url = urljoin(status_api.base_url, "kind/foo/resources/")

        response = requests.get(url)
        output = response.json()

        assert response.status_code == 200
        assert output == []

    def test_resources_add_resource(
        self,
        status_api: test_utils.RestServiceTestCase,
    ):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(
            uuid=uuid,
            name="foo-name",
            project_id=uuid,
        ).to_ua_resource(kind="/v1/kind/foo")
        view = resource.dump_to_simple_view()

        url = urljoin(status_api.base_url, "kind/foo/resources/")

        response = requests.post(url, json=view)
        output = response.json()

        assert response.status_code == 201
        assert output["uuid"] == str(uuid)

    def test_resources_update_resource(
        self,
        status_api: test_utils.RestServiceTestCase,
    ):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(
            uuid=uuid,
            name="foo-name",
            project_id=uuid,
        ).to_ua_resource(kind="/v1/kind/foo")
        view = resource.dump_to_simple_view()

        url = urljoin(status_api.base_url, "kind/foo/resources/")

        response = requests.post(url, json=view)
        output = response.json()

        assert response.status_code == 201
        assert output["uuid"] == str(uuid)

        view["value"]["name"] = "foo-name-updated"
        view.pop("uuid")
        view.pop("res_uuid")
        view.pop("created_at")
        view.pop("updated_at")

        url = urljoin(status_api.base_url, f"kind/foo/resources/{uuid}")

        response = requests.put(url, json=view)
        output = response.json()

        assert response.status_code == 200
        assert output["value"]["name"] == "foo-name-updated"

    def test_resources_delete_resource(
        self,
        status_api: test_utils.RestServiceTestCase,
    ):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(
            uuid=uuid,
            name="foo-name",
            project_id=uuid,
        ).to_ua_resource(kind="/v1/kind/foo")
        view = resource.dump_to_simple_view()

        url = urljoin(status_api.base_url, "kind/foo/resources/")

        response = requests.post(url, json=view)
        output = response.json()

        assert response.status_code == 201
        assert output["uuid"] == str(uuid)

        url = urljoin(status_api.base_url, f"kind/foo/resources/{uuid}")

        response = requests.delete(url)

        assert response.status_code == 204

    def test_resources_add_resource_dynamic_kind(
        self,
        status_api: test_utils.RestServiceTestCase,
    ):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(
            uuid=uuid,
            name="foo-name",
            project_id=uuid,
        ).to_ua_resource(kind="/v1/kind/foo")
        view = resource.dump_to_simple_view()

        url = urljoin(status_api.base_url, "kind/foo/resources/")

        response = requests.post(url, json=view)
        assert response.status_code == 201

        url = urljoin(status_api.base_url, "kind/")

        response = requests.get(url)
        output = response.json()

        assert response.status_code == 200
        assert output == ["foo"]

    def test_resources_list_kind(
        self,
        status_api: test_utils.RestServiceTestCase,
    ):
        uuid = sys_uuid.uuid4()
        resource = conftest.FooResource(
            uuid=uuid,
            name="foo-name",
            project_id=uuid,
        ).to_ua_resource(kind="/v1/kind/foo")
        view = resource.dump_to_simple_view()

        url = urljoin(status_api.base_url, "kind/foo/resources/")

        response = requests.post(url, json=view)
        assert response.status_code == 201

        response = requests.get(url)
        output = response.json()

        assert response.status_code == 200
        assert len(output) == 1
        assert output[0]["uuid"] == str(uuid)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import os
import socket
import contextlib
from urllib import parse

from restalchemy.storage.sql import migrations
from restalchemy.tests.functional import db_utils as ra_db_utils
from restalchemy.tests.functional.restapi.ra_based.microservice import service


ENDPOINT_TEMPLATE = "http://127.0.0.1:%s/"


class RestServiceTestCase(ra_db_utils.DBEngineMixin):
    __LAST_MIGRATION__ = None
    __FIRST_MIGRATION__ = None
    __API_VERSION__ = "v1"
    __APP__ = None

    @classmethod
    def setup_class(cls):
        cls.init_engine()

    @classmethod
    def teardown_class(cls):
        cls.drop_all_views()
        cls.drop_all_tables(cascade=True)
        cls.destroy_engine()

    @staticmethod
    def get_migration_engine(migrations_path: str | None = None) -> None:
        if migrations_path is None:
            migrations_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "../../migrations",
            )

        migration_engine = migrations.MigrationEngine(
            migrations_path=migrations_path
        )
        return migration_engine

    @property
    def base_url(self) -> str:
        return self.get_endpoint() + self.__API_VERSION__ + "/"

    @classmethod
    def drop_table(cls, table_name, session=None, cascade=False):
        cascade = " CASCADE" if cascade else ""
        with cls.engine.session_manager(session=session) as s:
            s.execute(
                "drop table if exists"
                f" {session.engine.escape(table_name)}{cascade}"
            )

    @classmethod
    def drop_all_tables(cls, session=None, cascade=False):
        with cls.engine.session_manager(session=session) as s:
            tables = cls.get_all_tables(session=s)
            for table in tables:
                cls.drop_table(table, session=s, cascade=cascade)

    @classmethod
    def get_all_views(cls, session=None) -> set[str]:
        with cls.engine.session_manager(session=session) as s:
            if session.engine.dialect.name == "mysql":
                res = s.execute(
                    """
                    select
                        table_name as table_name
                    from information_schema.views
                    where table_schema = database();
                """
                ).fetchall()
            elif session.engine.dialect.name == "postgresql":
                res = s.execute(
                    """
                    select
                        table_name as table_name
                    from information_schema.views
                    where table_schema = current_schema();
                """
                ).fetchall()
            else:
                raise NotImplementedError("Unsupported dialect")
        return {row["table_name"] for row in res}

    @classmethod
    def drop_all_views(cls, session=None):
        with cls.engine.session_manager(session=session) as s:
            views = cls.get_all_views(session=s)
            for view in views:
                cls.drop_view(view, session=s)

    @classmethod
    def drop_view(cls, view_name, session=None):
        with cls.engine.session_manager(session=session) as s:
            s.execute(
                f"drop view if exists {session.engine.escape(view_name)}"
            )

    def get_endpoint(self, template: str = ENDPOINT_TEMPLATE) -> str:
        return template % self.service_port

    def find_free_port(self) -> int:
        with contextlib.closing(
            socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        ) as s:
            s.bind(("127.0.0.1", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    def setup_method(self) -> None:
        # Apply migrations
        self._migrations = self.get_migration_engine()
        self._migrations.rollback_migration(self.__FIRST_MIGRATION__)

        last_migration = (
            self.__LAST_MIGRATION__ or self._migrations.get_latest_migration()
        )
        self._migrations.apply_migration(last_migration)

        # Run service
        self.service_port = self.find_free_port()
        url = parse.urlparse(self.get_endpoint())
        self._service = service.RESTService(
            bind_host=url.hostname, bind_port=url.port, app_root=self.__APP__
        )
        self._service.start()

    def teardown_method(self) -> None:
        self._service.stop()

        # Rollback migrations
        self._migrations = self.get_migration_engine()
        self._migrations.rollback_migration(self.__FIRST_MIGRATION__)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import typing as tp

from bazooka import exceptions as bazooka_exc

from gcl_sdk.clients.http import base as http
from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal.clients.backend import base
from gcl_sdk.agents.universal.clients.backend import exceptions


class RestApiBackendClient(base.AbstractBackendClient):
    """Rest API backend client."""

    def __init__(
        self,
        http_client: http.CollectionBaseClient,
        collection_map: dict[str, str],
    ) -> None:
        self._client = http_client
        self._collection_map = collection_map

    def get(self, resource: models.Resource) -> dict[str, tp.Any]:
        """Get the resource value in dictionary format."""
        collection_url = self._collection_map[resource.kind]

        try:
            return self._client.get(collection_url, resource.uuid)
        except bazooka_exc.NotFoundError:
            raise exceptions.ResourceNotFound(resource=resource)

    def create(self, resource: models.Resource) -> dict[str, tp.Any]:
        """Creates the resource. Returns the created resource."""
        collection_url = self._collection_map[resource.kind]

        try:
            return self._client.create(collection_url, resource.value)
        except bazooka_exc.ConflictError:
            raise exceptions.ResourceAlreadyExists(resource=resource)

    def update(self, resource: models.Resource) -> dict[str, tp.Any]:
        """Update the resource. Returns the updated resource."""
        collection_url = self._collection_map[resource.kind]

        try:
            return self._client.update(
                collection_url, resource.uuid, **resource.value
            )
        except bazooka_exc.NotFoundError:
            raise exceptions.ResourceNotFound(resource=resource)

    def list(self, kind: str, **kwargs) -> list[dict[str, tp.Any]]:
        """Lists all resources by kind."""
        collection_url = self._collection_map[kind]

        # TODO(akremenetsky): Use a project prefix to filter resources
        return self._client.filter(collection_url, **kwargs)

    def delete(self, resource: models.Resource) -> None:
        """Delete the resource."""
        collection_url = self._collection_map[resource.kind]

        try:
            self._client.delete(collection_url, resource.uuid)
        except bazooka_exc.NotFoundError:
            raise exceptions.ResourceNotFound(resource=resource)



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import logging
import itertools

from restalchemy.common import contexts
from restalchemy.dm import filters as dm_filters
from gcl_looper.services import basic as looper_basic

from gcl_sdk.agents.universal.dm import models
from gcl_sdk.agents.universal import constants as c


LOG = logging.getLogger(__name__)


class UniversalAgentSchedulerService(looper_basic.BasicService):

    def __init__(
        self,
        capabilities: list[str],
        iter_min_period=3,
        iter_pause=0.1,
    ):
        super().__init__(iter_min_period, iter_pause)
        self._capabilities = capabilities

    def _get_unscheduled_resources(
        self, limit: int = c.DEF_SQL_LIMIT
    ) -> list[models.TargetResource]:
        capabilities_in = []
        filters_in = []
        filters_like = []
        filters = []

        # Prepare like and in filters based on capabilities
        for cap in self._capabilities:
            if cap.endswith("*"):
                cap = cap.replace("*", "%")
                filters_like.append({"kind": dm_filters.Like(cap)})
            else:
                capabilities_in.append(cap)

        if capabilities_in:
            filters_in = [{"kind": dm_filters.In(capabilities_in)}]

        if not filters_like and not filters_in:
            LOG.debug("No filters for target resources")
            return

        filters = dm_filters.AND(
            {"agent": dm_filters.Is(None)},
            dm_filters.OR(*itertools.chain(filters_in + filters_like)),
        )

        return models.TargetResource.objects.get_all(
            filters=filters,
            limit=limit,
        )

    def _iteration(self):
        with contexts.Context().session_manager():
            unscheduled = self._get_unscheduled_resources()
            if not unscheduled:
                return

            LOG.info("Found %d unscheduled resources", len(unscheduled))

            cap_agent_map = models.UniversalAgent.have_capabilities(
                tuple(u.kind for u in unscheduled)
            )

            # FIXME(akremenetsky): The simplest implementation.
            # Take a first agent that satisfies
            for resource in unscheduled:
                if resource.kind not in cap_agent_map:
                    LOG.warning("No agent for resource %s", resource.kind)
                    continue

                # It's supposed only one agent have a particular EM capability
                # In case of multiple agents, the first one is chosen for nodes,
                # the second one is chosen for configs and so on.
                agent = cap_agent_map[resource.kind][0]
                resource.agent = agent.uuid
                resource.node = agent.node
                resource.save()
                LOG.info(
                    "Assigning resource %s to agent %s",
                    resource.uuid,
                    agent.uuid,
                )



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os
import enum


GENESIS_WORK_DIR = "/var/lib/genesis"
WORK_DIR = "/var/lib/genesis/universal_agent/"
PAYLOAD_PATH = os.path.join(WORK_DIR, "payload.json")
NODE_UUID_PATH = os.path.join(GENESIS_WORK_DIR, "node-id")
DEFAULT_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%1f"
EP_UNIVERSAL_AGENT = "gcl_sdk_universal_agent"

DEF_SQL_LIMIT = 100


class AgentStatus(str, enum.Enum):
    NEW = "NEW"
    ACTIVE = "ACTIVE"
    ERROR = "ERROR"
    DISABLED = "DISABLED"


class InstanceStatus(str, enum.Enum):
    NEW = "NEW"
    IN_PROGRESS = "IN_PROGRESS"
    ACTIVE = "ACTIVE"
    ERROR = "ERROR"



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from oslo_config import cfg

from gcl_sdk.common import constants
from gcl_sdk import version


GLOBAL_SERVICE_NAME = constants.GLOBAL_SERVICE_NAME


_CONFIG_NOT_FOUND_MESSAGE = (
    "Unable to find configuration file in the"
    " default search paths (~/.%(service_name)s/, ~/,"
    " /etc/%(service_name)s/, /etc/) and the '--config-file' option!"
    % {"service_name": GLOBAL_SERVICE_NAME}
)


def parse(args):
    cfg.CONF(
        args=args,
        project=GLOBAL_SERVICE_NAME,
        version="%s %s"
        % (
            GLOBAL_SERVICE_NAME.capitalize(),
            version.version_info,
        ),
    )
    if not cfg.CONF.config_file:
        logging.warning(_CONFIG_NOT_FOUND_MESSAGE)
    return cfg.CONF.config_file



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

import logging
import uuid as sys_uuid
import typing as tp

from gcl_sdk.agents.universal.dm import models as ua_models
from gcl_sdk.agents.universal.services import builder


LOG = logging.getLogger(__name__)


class PaaSResourcePair(tp.NamedTuple):
    target: ua_models.TargetResourceKindAwareMixin
    actual: ua_models.TargetResourceKindAwareMixin | None


class PaaSCollection(tp.NamedTuple):
    paas_objects: tp.Collection[PaaSResourcePair]

    def targets(self) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        return tuple(p.target for p in self.paas_objects)

    def actuals(self) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        return tuple(p.actual for p in self.paas_objects)


class PaaSBuilder(builder.UniversalBuilderService):

    def create_paas_objects(
        self, instance: ua_models.InstanceWithDerivativesMixin
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Create a list of PaaS objects.

        The method returns a list of PaaS objects that are required
        for the instance.
        """
        raise NotImplementedError

    def actualize_paas_objects(
        self,
        instance: ua_models.InstanceWithDerivativesMixin,
        paas_collection: PaaSCollection,
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Update a list of PaaS objects.

        The method processes existing resources, actualize them
        if needed, creates new objects if needed, and returns only actual
        and updated PaaS objects for the instance.

        Basic implementation ignores existing objects and just recreates them.
        """
        return self.create_paas_objects(instance)

    def schedule_paas_objects(
        self,
        instance: ua_models.InstanceWithDerivativesMixin,
        paas_objects: tp.Collection[ua_models.TargetResourceKindAwareMixin],
    ) -> dict[
        sys_uuid.UUID, tp.Collection[ua_models.TargetResourceKindAwareMixin]
    ]:
        """Schedule the PaaS objects.

        The method schedules the PaaS objects. The result is a dictionary
        where the key is a UUID of a agent and the value is a list of PaaS
        objects that should be scheduled on this agent.
        """
        return {}

    def enable_schedule_paas_objects(self) -> bool:
        """Enable schedule PaaS objects."""
        return True

    def actualize_paas_objects_source_data_plane(
        self,
        instance: ua_models.InstanceWithDerivativesMixin,
        paas_collection: PaaSCollection,
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Actualize the PaaS objects. Changes from the data plane.

        The method is called when the instance is outdated. For example,
        the instance `Database` has derivative `PGDatabase`. Single `Database`
        may have multiple `PGDatabase` derivatives. If any of the derivatives
        is outdated, this method is called to reactualize this PaaS objects.

        Args:
            instance: The instance to actualize.
            paas_objects: The actual PaaS objects.
        """
        return self.actualize_paas_objects(instance, paas_collection)

    def actualize_paas_objects_source_master(
        self,
        instance: ua_models.InstanceWithDerivativesMixin,
        master_instance: ua_models.InstanceWithDerivativesMixin,
        paas_collection: PaaSCollection,
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Actualize the PaaS objects. Changes from the master instance.

        The method is called when the instance is outdated from master
        instance point of view. For example, the instance `Database` is linked to the
        `NodeSet` instance. If the `NodeSet` is outdated, this method is called
        to reactualize the `Database` instance.

        Args:
            instance: The instance to actualize.
            master_instance: The master instance.
            paas_collection: The actual PaaS objects.
        """
        return paas_collection.targets()

    def create_instance_derivatives(
        self, instance: ua_models.InstanceWithDerivativesMixin
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Create the instance.

        The result is an iterable object of derivative objects that are
        required for the instance. For example, the main instance is a
        `Config` so the derivative objects for the the config is a list
        of `Render`. The result is a tuple/list/set/... of render objects.
        The derivative objects should inherit from the `TargetResourceMixin`.
        """
        return self.create_paas_objects(instance)

    def update_instance_derivatives(
        self,
        instance: ua_models.InstanceWithDerivativesMixin,
        resource: ua_models.TargetResourceKindAwareMixin,
        derivative_pairs: tp.Collection[
            tuple[
                ua_models.TargetResourceKindAwareMixin,  # The target resource
                ua_models.TargetResourceKindAwareMixin
                | None,  # The actual resource
            ]
        ],
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """The hook to update instance derivatives.

        The hook is called when an initiator of updating is an user or
        software from control plane side.
        The default behavior is to send the same list as on instance creation.
        """
        paas_collection = PaaSCollection(
            paas_objects=tuple(PaaSResourcePair(*p) for p in derivative_pairs),
        )

        return self.actualize_paas_objects(instance, paas_collection)

    def actualize_outdated_instance_derivatives(
        self,
        instance: ua_models.InstanceWithDerivativesMixin,
        derivative_pairs: tp.Collection[
            tuple[
                ua_models.TargetResourceKindAwareMixin,  # The target resource
                ua_models.TargetResourceKindAwareMixin
                | None,  # The actual resource
            ]
        ],
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Actualize outdated instance with derivatives.

        It means some changes occurred on the data plane and the instance
        is outdated now. For example, the instance `Config` has derivative
        `Render`. Single `Config` may have multiple `Render` derivatives.
        If any of the derivatives is outdated, this method is called to
        reactualize the derivatives. The method returns the list of `updated`
        derivatives. If nothing needs to be updated, the method returns the
        same list of target derivatives as it received. Otherwise, the method
        should return the list of updated derivatives. It also can add new or
        remove old derivatives.
        Depends on the `fetch_all_derivatives_on_outdate` the behavior of the
        method is different:

        fetch_all_derivatives_on_outdate == True:
        The method receives the list of all derivatives currently available
        for the instance even though the derivatives aren't outdated.

        fetch_all_derivatives_on_outdate == False:
        The method receives the list only changed derivatives from the last
        actualization. For example, a config has two renders. Only one of
        them is outdated. The method receives the list of only one outdated
        render.

        Args:
            instance: The instance to actualize.
            derivative_pairs: Changed or all derivatives of the instance.
        """
        paas_collection = PaaSCollection(
            paas_objects=tuple(PaaSResourcePair(*p) for p in derivative_pairs),
        )

        return self.actualize_paas_objects_source_data_plane(
            instance, paas_collection
        )

    def track_outdated_master_full_hash_instances(self) -> bool:
        """Track outdated master full hash instances."""
        return True

    def actualize_outdated_master_full_hash_instance(
        self,
        instance: ua_models.InstanceWithDerivativesMixin,
        master_instance: ua_models.InstanceWithDerivativesMixin,
        derivatives: tp.Collection[
            tuple[
                ua_models.TargetResourceKindAwareMixin,  # The target resource
                ua_models.TargetResourceKindAwareMixin
                | None,  # The actual resource
            ]
        ],
    ) -> tp.Collection[ua_models.TargetResourceKindAwareMixin]:
        """Actualize outdated master full hash instance.

        The logic is quite similar to `actualize_outdated_instance_derivatives`.
        But the reason when this method is called is different. The
        `actualize_outdated_instance_derivatives` allows to track changes on the
        data plane but this method allows to track changes on a related master
        instance. For example, the instance model is `Database`, the related master
        for this instance is `NodeSet`. If the `NodeSet` is updated, this method
        is called for all `Database` instances that are related to this `NodeSet`
        to reactualize them.
        This method tracks changes for all fields of the master instance.

        Args:
            instance: The instance to actualize.
            master_instance: The master instance.
            derivatives: All derivatives of the instance.
        """
        paas_collection = PaaSCollection(
            paas_objects=tuple(PaaSResourcePair(*p) for p in derivatives),
        )

        return self.actualize_paas_objects_source_master(
            instance, master_instance, paas_collection
        )

    def post_create_instance_resource(
        self,
        instance: (
            ua_models.InstanceMixin | ua_models.InstanceWithDerivativesMixin
        ),
        resource: ua_models.TargetResource,
        derivatives: tp.Collection[ua_models.TargetResource] = tuple(),
    ) -> None:
        """The hook is performed after saving instance resource.

        The hook is called only for new instances.
        """
        super().post_create_instance_resource(instance, resource, derivatives)

        if not self.enable_schedule_paas_objects():
            return

        # Prepare PaaS objects for scheduling
        paas_objects = []
        for derivative in derivatives:
            class_ = self._instance_model.derivative_model(derivative.kind)
            paas_objects.append(class_.from_ua_resource(derivative))

        schedule_map = self.schedule_paas_objects(
            instance,
            paas_objects,
        )

        # Schedule PaaS objects
        # We don't expect to have a lot of derivatives.
        # So it's not a big problem to have nested loop here
        for agent_uuid, paas_objects in schedule_map.items():
            for paas_object in paas_objects:
                for derivative in derivatives:
                    if derivative.uuid == paas_object.uuid:
                        derivative.agent = agent_uuid



#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import annotations

from restalchemy.dm import models
from restalchemy.dm import properties
from restalchemy.dm import filters as dm_filters
from restalchemy.dm import types
from restalchemy.dm import models as ra_models
from restalchemy.dm import relationships
from restalchemy.storage.sql import engines
from restalchemy.storage import base as base_storage
from restalchemy.common import utils as ra_utils
from restalchemy.storage import exceptions as storage_exc

from gcl_sdk.agents.universal.dm import models


class KindCollection(base_storage.AbstractObjectCollection):
    def get_all(
        self,
        filters: dict[str, dm_filters.AbstractClause] | None = None,
        limit: int | None = None,
        order_by: dict | None = None,
        locked: bool = False,
    ) -> list[str]:
        expression = "SELECT kind FROM ua_actual_resources GROUP BY kind;"

        engine = engines.engine_factory.get_engine()
        with engine.session_manager() as session:
            curs = session.execute(expression, tuple())
            response = curs.fetchall()

        return [k["kind"] for k in response]

    def get_one(
        self,
        filters: dict[str, dm_filters.AbstractClause] | None = None,
        locked: bool = False,
    ) -> KindModel:
        # Unable get kind, means the collection cannot be found
        try:
            kind = filters["kind"].value
        except Exception:
            raise storage_exc.RecordNotFound(
                model=self.model_cls, filters=filters
            )

        return self.model_cls(kind=kind)


class KindModel(ra_models.ModelWithID):

    kind = properties.property(
        types.String(max_length=64),
        required=True,
        read_only=True,
        id_property=True,
    )

    @ra_utils.classproperty
    def objects(cls):
        return KindCollection(cls)

    @classmethod
    def to_simple_type(cls, value: KindModel | str) -> str:
        if isinstance(value, KindModel):
            return value.kind
        return value

    @classmethod
    def from_simple_type(self, value: str) -> KindModel:
        return KindModel(kind=value)


class Resource(models.Resource):
    kind = relationships.relationship(KindModel, required=True)


# NOTE(akremenetsky): Keep it as a separate class to be able to
# launch functional tests
class UniversalAgent(models.UniversalAgent):
    pass
